#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import numpy

import uproot3_methods.base


class Methods(uproot3_methods.base.ROOTMethods):
    @property
    def numbins(self):
        return self.xnumbins * self.ynumbins

    @property
    def xnumbins(self):
        return self._fXaxis._fNbins

    @property
    def ynumbins(self):
        return self._fYaxis._fNbins

    @property
    def low(self):
        return self.xlow, self.ylow

    @property
    def high(self):
        return self.xhigh, self.yhigh

    @property
    def xlow(self):
        return self._fXaxis._fXmin

    @property
    def xhigh(self):
        return self._fXaxis._fXmax

    @property
    def ylow(self):
        return self._fYaxis._fXmin

    @property
    def yhigh(self):
        return self._fYaxis._fXmax

    @property
    def underflows(self):
        uf = numpy.array(self.allvalues)
        xuf = uf[0]
        yuf = uf[:, 0]
        return xuf, yuf

    @property
    def xunderflows(self):
        return self.underflows[0]

    @property
    def yunderflows(self):
        return self.underflows[1]

    @property
    def overflows(self):
        of = numpy.array(self.allvalues)
        xof = of[-1]
        yof = of[:, -1]
        return xof, yof

    @property
    def xoverflows(self):
        return self.overflows[0]

    @property
    def yoverflows(self):
        return self.overflows[1]

    @property
    def edges(self):
        xaxis = self._fXaxis
        yaxis = self._fYaxis
        if len(getattr(xaxis, "_fXbins", [])) > 0:
            xedges = numpy.array(xaxis._fXbins)
        else:
            xedges = numpy.linspace(xaxis._fXmin, xaxis._fXmax, xaxis._fNbins + 1)
        if len(getattr(yaxis, "_fXbins", [])) > 0:
            yedges = numpy.array(yaxis._fXbins)
        else:
            yedges = numpy.linspace(yaxis._fXmin, yaxis._fXmax, yaxis._fNbins + 1)
        return xedges, yedges

    @property
    def alledges(self):
        xedges, yedges = self.edges
        vx = numpy.empty(len(xedges) + 2)
        vy = numpy.empty(len(yedges) + 2)
        vx[0] = -numpy.inf
        vx[-1] = numpy.inf
        vx[1:-1] = xedges
        vy[0] = -numpy.inf
        vy[-1] = numpy.inf
        vy[1:-1] = yedges
        return vx, vy

    @property
    def bins(self):
        xedges, yedges = self.edges
        vx = numpy.empty((len(xedges) - 1, 2))
        vy = numpy.empty((len(yedges) - 1, 2))
        vx[:, 0] = xedges[:-1]
        vx[:, 1] = xedges[1:]
        vy[:, 0] = yedges[:-1]
        vy[:, 1] = yedges[1:]
        return vx, vy

    @property
    def allbins(self):
        xedges, yedges = self.alledges
        vx = numpy.empty((len(xedges) - 1, 2))
        vy = numpy.empty((len(yedges) - 1, 2))
        vx[:, 0] = xedges[:-1]
        vx[:, 1] = xedges[1:]
        vy[:, 0] = yedges[:-1]
        vy[:, 1] = yedges[1:]
        return vx, vy

    @property
    def values(self):
        va = self.allvalues
        return va[1:self.xnumbins+1, 1:self.ynumbins+1]

    @property
    def allvalues(self):
        dtype = getattr(self, "_dtype", numpy.dtype(numpy.float64))
        v = numpy.array(self[:], dtype=dtype.newbyteorder("="))
        v = v.reshape(self.ynumbins + 2, self.xnumbins + 2)
        return v.T

    @property
    def variances(self):
        va = self.allvariances
        return va[1:self.xnumbins+1, 1:self.ynumbins+1]

    @property
    def allvariances(self):
        if len(getattr(self, "_fSumw2", [])) != len(self):
            v = numpy.array(self, dtype=numpy.float64)
        else:
            v = numpy.array(self._fSumw2, dtype=numpy.float64)
        v = v.reshape(self.ynumbins + 2, self.xnumbins + 2)
        return v.T

    def numpy(self):
        return (self.values, [self.edges])

    def allnumpy(self):
        return (self.allvalues, [self.alledges])

    def interval(self, index, axis):
        if axis == "x":
            low = self.xlow
            high = self.xhigh
            nbins = self.xnumbins
            bins = self._fXaxis._fXbins
        elif axis == "y":
            low = self.ylow
            high = self.yhigh
            nbins = self.ynumbins
            bins = self._fYaxis._fXbins
        else:
            raise ValueError("axis must be 'x' or 'y'")

        if index < 0:
            index += nbins

        if index == 0:
            return float("-inf"), low
        elif index == nbins + 1:
            return high, float("inf")
        else:
            if not bins:
                norm = float(high-low) / nbins
                xedges = (index-1)*norm + low, index*norm + low
            else:
                xedges = bins[index-1], bins[index]
            return xedges

    def xinterval(self, index):
        return self.interval(index, "x")

    def yinterval(self, index):
        return self.interval(index, "y")

    @property
    def ylabels(self):
        if self._fYaxis._fLabels is None:
            return None
        else:
            return [str(x) for x in self._fYaxis._fLabels]

def _histtype(content):
    if issubclass(content.dtype.type, (numpy.bool_, numpy.bool)):
        return b"TH2C", content.astype(">i1")
    elif issubclass(content.dtype.type, numpy.int8):
        return b"TH2C", content.astype(">i1")
    elif issubclass(content.dtype.type, numpy.uint8) and content.max() <= numpy.iinfo(numpy.int8).max:
        return b"TH2C", content.astype(">i1")
    elif issubclass(content.dtype.type, numpy.uint8):
        return b"TH2S", content.astype(">i2")
    elif issubclass(content.dtype.type, numpy.int16):
        return b"TH2S", content.astype(">i2")
    elif issubclass(content.dtype.type, numpy.uint16) and content.max() <= numpy.iinfo(numpy.int16).max:
        return b"TH2S", content.astype(">i2")
    elif issubclass(content.dtype.type, numpy.uint16):
        return b"TH2I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.int32):
        return b"TH2I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.uint32) and content.max() <= numpy.iinfo(numpy.int32).max:
        return b"TH2I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.integer) and numpy.iinfo(numpy.int32).min <= content.min() and content.max() <= numpy.iinfo(numpy.int32).max:
        return b"TH2I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.float32):
        return b"TH2F", content.astype(">f4")
    else:
        return b"TH2D", content.astype(">f8")

def from_numpy(histogram):
    if isinstance(histogram[1], list) and len(histogram[1]) == 2:
        content, (xedges, yedges) = histogram[:2]
    else:
        content, xedges, yedges = histogram[:3]

    class TH2(Methods, list):
        pass

    class TAxis(object):
        def __init__(self, edges):
            self._fNbins = len(edges) - 1
            self._fXmin = edges[0]
            self._fXmax = edges[-1]
            if numpy.array_equal(edges, numpy.linspace(self._fXmin, self._fXmax, len(edges), dtype=edges.dtype)):
                self._fXbins = numpy.array([], dtype=">f8")
            else:
                self._fXbins = edges.astype(">f8")

    out = TH2.__new__(TH2)
    out._fXaxis = TAxis(xedges)
    out._fYaxis = TAxis(yedges)
    out._fEntries = out._fTsumw = out._fTsumw2 = content.sum()

    xcenters = (xedges[:-1] + xedges[1:]) / 2.
    out._fTsumwx = xcenters.dot(content.sum(1))
    out._fTsumwx2 = (xcenters**2).dot(content.sum(1))

    ycenters = (yedges[:-1] + yedges[1:]) / 2.
    out._fTsumwy = ycenters.dot(content.sum(0))
    out._fTsumwy2 = (ycenters**2).dot(content.sum(0))

    out._fTsumwxy = xcenters.dot(content).dot(ycenters)

    if len(histogram) >= 4:
        out._fTitle = histogram[3]
    else:
        out._fTitle = b""

    out._classname, content = _histtype(content)

    valuesarray = numpy.pad(content.T, (1, 1), mode='constant').flatten()

    out.extend(valuesarray)
    out._fSumw2 = valuesarray**2

    return out



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import awkward0
import awkward0.util

class ROOTMethods(awkward0.Methods):
    _arraymethods = None

    awkward = awkward0
    awkward0 = awkward0

    def __ne__(self, other):
        return not self.__eq__(other)

    def _trymemo(self, name, function):
        memoname = "_memo_" + name
        wrap, (array,) = awkward0.util.unwrap_jagged(type(self), self.JaggedArray, (self,))
        if not hasattr(array, memoname):
            setattr(array, memoname, function(array))
        return wrap(getattr(array, memoname))



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import math
import numbers
import operator

import awkward0.array.chunked
import awkward0.array.jagged
import awkward0.array.objects
import awkward0.util

import uproot3_methods.base
import uproot3_methods.common.TVector
import uproot3_methods.classes.TVector3

class Common(object):
    @property
    def E(self):
        return self.t

    def dot(self, other):
        out = self.t*other.t
        out = out - self.x*other.x
        out = out - self.y*other.y
        out = out - self.z*other.z
        return out

    @property
    def energy(self):
        return self.t

    @property
    def p(self):
        return self.p3.mag

    @property
    def p2(self):
        return self.p3.mag2

    @property
    def perp2(self):
        return self.p3.rho2

    @property
    def perp(self):
        return self.p3.rho

    @property
    def pt2(self):
        return self.p3.rho2

    @property
    def Et(self):
        return self.energy * self.pt / self.p

    @property
    def mag2(self):
        return self.dot(self)

    @property
    def mass2(self):
        return self.mag2

    @property
    def mt2(self):
        return self.energy**2 - self.z**2

    @property
    def theta(self):
        return self.p3.theta

    @property
    def cottheta(self):
        return self.p3.cottheta

    @property
    def beta(self):
        return self.p / self.energy

    def delta_phi(self, other):
        return (self.phi - other.phi + math.pi) % (2*math.pi) - math.pi

    def delta_r2(self, other):
        return (self.eta - other.eta)**2 + self.delta_phi(other)**2

    def _rotate_axis(self, axis, angle):
        if not isinstance(axis, uproot3_methods.classes.TVector3.Common):
            raise TypeError("axis must be an (array of) TVector3")
        p3 = self.p3._rotate_axis(axis, angle)
        return p3, self.t

    def _rotate_euler(self, phi, theta, psi):
        return self.p3._rotate_euler(phi, theta, psi), self.t

    def rotatex(self, angle):
        return self.rotate_axis(uproot3_methods.classes.TVector3.TVector3(1.0, 0.0, 0.0), angle)

    def rotatey(self, angle):
        return self.rotate_axis(uproot3_methods.classes.TVector3.TVector3(0.0, 1.0, 0.0), angle)

    def rotatez(self, angle):
        return self.rotate_axis(uproot3_methods.classes.TVector3.TVector3(0.0, 0.0, 1.0), angle)

    def isspacelike(self, tolerance=1e-10):
        return self.mag2 < -tolerance

    def istimelike(self, tolerance=1e-10):
        return self.mag2 > tolerance

    def __lt__(self, other):
        raise TypeError("Lorentz vectors have no natural ordering")

    def __gt__(self, other):
        raise TypeError("Lorentz vectors have no natural ordering")

    def __le__(self, other):
        raise TypeError("Lorentz vectors have no natural ordering")

    def __ge__(self, other):
        raise TypeError("Lorentz vectors have no natural ordering")

class ArrayMethods(Common, uproot3_methods.base.ROOTMethods):
    def _initObjectArray(self, table):
        self.awkward0.ObjectArray.__init__(self, table, lambda row: TLorentzVector(row["fX"], row["fY"], row["fZ"], row["fE"]))

    def __awkward_serialize__(self, serializer):
        self._valid()
        x, y, z, t = self.x, self.y, self.z, self.t
        return serializer.encode_call(
            ["uproot3_methods.classes.TLorentzVector", "TLorentzVectorArray", "from_cartesian"],
            serializer(x, "TLorentzVectorArray.x"),
            serializer(y, "TLorentzVectorArray.y"),
            serializer(z, "TLorentzVectorArray.z"),
            serializer(t, "TLorentzVectorArray.t"))

    @staticmethod
    def _wrapmethods(node, awkwardlib):
        if isinstance(node, awkward0.array.chunked.ChunkedArray):
            node.__class__ = type("ChunkedArrayMethods", (awkwardlib.ChunkedArray, uproot3_methods.classes.TVector3.ArrayMethods), {})
            for chunk in node.chunks:
                ArrayMethods._wrapmethods(chunk, awkwardlib)
        elif isinstance(node, awkward0.array.jagged.JaggedArray):
            node.__class__ = type("JaggedArrayMethods", (awkwardlib.JaggedArray, uproot3_methods.classes.TVector3.ArrayMethods), {})
            ArrayMethods._wrapmethods(node.content, awkwardlib)
        elif isinstance(node, awkward0.array.objects.ObjectArray):
            node.__class__ = type("ObjectArrayMethods", (awkwardlib.ObjectArray, uproot3_methods.classes.TVector3.ArrayMethods), {})

    @property
    def p3(self):
        out = self.empty_like(generator=lambda row: uproot3_methods.classes.TVector3.TVector3(row["fX"], row["fY"], row["fZ"]))
        ArrayMethods._wrapmethods(out, self.awkward0)
        out["fX"] = self.x
        out["fY"] = self.y
        out["fZ"] = self.z
        return out

    @property
    def x(self):
        return self["fX"]

    @property
    def y(self):
        return self["fY"]

    @property
    def z(self):
        return self["fZ"]

    @property
    def t(self):
        return self["fE"]

    @property
    def pt(self):
        return self._trymemo("pt", lambda self: self.awkward0.numpy.sqrt(self.pt2))

    @property
    def eta(self):
        return self._trymemo("eta", lambda self: self.awkward0.numpy.arcsinh(self.z / self.awkward0.numpy.sqrt(self.x**2 + self.y**2)))

    @property
    def phi(self):
        return self._trymemo("phi", lambda self: self.p3.phi)

    @property
    def mass(self):
        return self._trymemo("mass", lambda self: self.awkward0.numpy.sqrt(self.mag2))

    @property
    def mag(self):
        return self.awkward0.numpy.sqrt(self.mag2)

    @property
    def mt(self):
        mt2 = self.mt2
        sign = self.awkward0.numpy.sign(mt2)
        return self.awkward0.numpy.sqrt(self.awkward0.numpy.absolute(mt2)) * sign

    @property
    def rapidity(self):
        return 0.5 * self.awkward0.numpy.log((self.t + self.z) / (self.t - self.z))

    @property
    def unit(self):
        return self / self.awkward0.numpy.sqrt(self.mag)

    @property
    def boostp3(self):
        out = self.empty_like(generator=lambda row: uproot3_methods.classes.TVector3.TVector3(row["fX"], row["fY"], row["fZ"]))
        if isinstance(self, self.awkward0.JaggedArray):
            out.__class__ = type("JaggedArrayMethods", (self.awkward0.JaggedArray, uproot3_methods.classes.TVector3.ArrayMethods), {})
        else:
            out.__class__ = type("ObjectArrayMethods", (self.awkward0.ObjectArray, uproot3_methods.classes.TVector3.ArrayMethods), {})
        out["fX"] = self.x / self.t
        out["fY"] = self.y / self.t
        out["fZ"] = self.z / self.t
        return out

    def boost(self, p3):
        if not isinstance(p3, (uproot3_methods.classes.TVector3.ArrayMethods, uproot3_methods.classes.TVector3.Methods)):
            raise TypeError("boost p3 must be an (array of) TVector3")

        b2 = p3.mag2
        gamma = (1 - b2)**(-0.5)
        gamma2 = self.awkward0.numpy.zeros(b2.shape, dtype=self.awkward0.numpy.float64)
        mask = (b2 != 0)
        gamma2[mask] = (gamma[mask] - 1) / b2[mask]
        del mask

        bp = self.p3.dot(p3)

        v = self.p3 + gamma2*bp*p3 + self.t*gamma*p3
        out = self.empty_like()
        out["fX"] = v.x
        out["fY"] = v.y
        out["fZ"] = v.z
        out["fE"] = gamma*(self.t + bp)
        return out

    @property
    def gamma(self):
        out = self.beta
        mask = (out < 1) & (out > -1)
        out[mask] = (1 - out[mask]**2)**(-0.5)
        out[~mask] = self.awkward0.numpy.inf
        return out

    def delta_r(self, other):
        return self.awkward0.numpy.sqrt(self.delta_r2(other))

    def rotate_axis(self, axis, angle):
        p3, t = self._rotate_axis(axis, angle)
        x, y, z = p3
        out = self.empty_like()
        out["fX"] = x
        out["fY"] = y
        out["fZ"] = z
        out["fE"] = t
        return out

    def rotate_euler(self, phi=0, theta=0, psi=0):
        p3, t = self._rotate_euler(phi, theta, psi)
        x, y, z = p3
        out = self.empty_like()
        out["fX"] = x
        out["fY"] = y
        out["fZ"] = z
        out["fE"] = t
        return out

    def islightlike(self, tolerance=1e-10):
        return self.awkward0.numpy.absolute(self.mag2) < tolerance

    def sum(self):
        if isinstance(self, awkward0.AwkwardArray) and self._util_hasjagged(self):
            return TLorentzVectorArray.from_cartesian(self.x.sum(), self.y.sum(), self.z.sum(), self.t.sum())
        else:
            return TLorentzVector(self.x.sum(), self.y.sum(), self.z.sum(), self.t.sum())

    def _to_cartesian(self):
        return TLorentzVectorArray.from_cartesian(self.x,self.y,self.z,self.t)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if "out" in kwargs:
            raise NotImplementedError("in-place operations not supported")

        if method != "__call__":
            return NotImplemented

        inputs = list(inputs)
        for i in range(len(inputs)):
            if isinstance(inputs[i], self.awkward0.numpy.ndarray) and inputs[i].dtype == self.awkward0.numpy.dtype(object) and len(inputs[i]) > 0:
                idarray = self.awkward0.numpy.frombuffer(inputs[i], dtype=self.awkward0.numpy.uintp)
                if (idarray == idarray[0]).all():
                    inputs[i] = inputs[i][0]

        if ufunc is self.awkward0.numpy.add or ufunc is self.awkward0.numpy.subtract:
            if not all(isinstance(x, (ArrayMethods, Methods)) for x in inputs):
                raise TypeError("(arrays of) TLorentzVector can only be added to/subtracted from other (arrays of) TLorentzVector")
            cart_inputs = [x._to_cartesian() for x in inputs]
            out = cart_inputs[0].empty_like()
            out["fX"] = getattr(ufunc, method)(*[x.x for x in cart_inputs], **kwargs)
            out["fY"] = getattr(ufunc, method)(*[x.y for x in cart_inputs], **kwargs)
            out["fZ"] = getattr(ufunc, method)(*[x.z for x in cart_inputs], **kwargs)
            out["fE"] = getattr(ufunc, method)(*[x.t for x in cart_inputs], **kwargs)
            return out

        elif ufunc is self.awkward0.numpy.power and len(inputs) >= 2 and isinstance(inputs[1], (numbers.Number, self.awkward0.numpy.number)):
            if inputs[1] == 2:
                return self.mag2
            else:
                return self.mag2**(0.5*inputs[1])

        elif ufunc is self.awkward0.numpy.absolute:
            return self.mag

        else:
            return super(ArrayMethods, self).__array_ufunc__(ufunc, method, *inputs, **kwargs)

JaggedArrayMethods = ArrayMethods.mixin(ArrayMethods, awkward0.JaggedArray)

class PtEtaPhiMassArrayMethods(ArrayMethods):
    def _initObjectArray(self, table):
        self.awkward0.ObjectArray.__init__(self, table, lambda row: PtEtaPhiMassLorentzVector(row["fPt"], row["fEta"], row["fPhi"], row["fMass"]))

    def __awkward_serialize__(self, serializer):
        self._valid()
        pt, eta, phi, mass = self.pt, self.eta, self.phi, self.mass
        return serializer.encode_call(
            ["uproot3_methods.classes.TLorentzVector", "TLorentzVectorArray", "from_ptetaphim"],
            serializer(pt, "TLorentzVectorArray.pt"),
            serializer(eta, "TLorentzVectorArray.eta"),
            serializer(phi, "TLorentzVectorArray.phi"),
            serializer(mass, "TLorentzVectorArray.mass"))

    @property
    def x(self):
        return self._trymemo("x",lambda self: self.pt * self.awkward0.numpy.cos(self.phi))

    @property
    def y(self):
        return self._trymemo("y",lambda self: self.pt * self.awkward0.numpy.sin(self.phi))

    @property
    def z(self):
        return self._trymemo("z",lambda self: self.pt * self.awkward0.numpy.sinh(self.eta))

    @property
    def t(self):
        return self._trymemo("t",lambda self: self.awkward0.numpy.hypot(self.mass, self.p))

    @property
    def pt(self):
        return self["fPt"]

    @property
    def pt2(self):
        return self["fPt"]**2

    @property
    def perp(self):
        return self["fPt"]

    @property
    def perp2(self):
        return self["fPt"]**2

    @property
    def eta(self):
        return self["fEta"]

    @property
    def phi(self):
        return self["fPhi"]

    @property
    def mass(self):
        return self["fMass"]

    @property
    def mass2(self):
        return self["fMass"]**2

    @property
    def mag(self):
        return self["fMass"]

    @property
    def mag2(self):
        return self["fMass"]**2

    @property
    def mt(self):
        return self.awkward0.numpy.sqrt(self.mt2)

    @property
    def mt2(self):
        return self["fMass"]**2 + self["fPt"]**2

    @property
    def p(self):
        return self._trymemo("p",lambda self: self["fPt"]*self.awkward0.numpy.cosh(self["fEta"]))

    @property
    def p2(self):
        return self.p**2

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if "out" in kwargs:
            raise NotImplementedError("in-place operations not supported")

        if method != "__call__":
            return NotImplemented

        inputs = list(inputs)
        for i in range(len(inputs)):
            if isinstance(inputs[i], self.awkward0.numpy.ndarray) and inputs[i].dtype == self.awkward0.numpy.dtype(object) and len(inputs[i]) > 0:
                idarray = self.awkward0.numpy.frombuffer(inputs[i], dtype=self.awkward0.numpy.uintp)
                if (idarray == idarray[0]).all():
                    inputs[i] = inputs[i][0]

        if ufunc is self.awkward0.numpy.multiply or ufunc is self.awkward0.numpy.divide:
            if sum(isinstance(x, PtEtaPhiMassArrayMethods) for x in inputs) > 1:
                raise ValueError("cannot multiply or divide two PtEtaPhiMassArrayMethods")
            this_input = None
            for i in range(len(inputs)):
                if isinstance(inputs[i], PtEtaPhiMassArrayMethods) and not isinstance(inputs[i], self.awkward0.JaggedArray) and this_input is None:
                    this_input = inputs[i]
                    inputs[i] = self.awkward0.Table(fPt=inputs[i]['fPt'], fMass=inputs[i]['fMass'])

            out = super(PtEtaPhiMassArrayMethods, self).__array_ufunc__(ufunc, method, *inputs, **kwargs)
            if this_input is not None:
                out['fEta'] = this_input['fEta']
                out['fPhi'] = this_input['fPhi']
                out.__class__ = this_input.__class__
            return out

        else:
            return super(PtEtaPhiMassArrayMethods, self).__array_ufunc__(ufunc, method, *inputs, **kwargs)

PtEtaPhiMassJaggedArrayMethods = PtEtaPhiMassArrayMethods.mixin(PtEtaPhiMassArrayMethods, awkward0.JaggedArray)

class Methods(Common, uproot3_methods.base.ROOTMethods):
    _arraymethods = ArrayMethods

    @property
    def p3(self):
        return self._fP

    @property
    def x(self):
        return self._fP._fX

    @property
    def y(self):
        return self._fP._fY

    @property
    def z(self):
        return self._fP._fZ

    @property
    def t(self):
        return self._fE

    def _to_cartesian(self):
        return TLorentzVector(self.x,self.y,self.z,self.t)

    def __repr__(self):
        return "TLorentzVector(x={0:.5g}, y={1:.5g}, z={2:.5g}, t={3:.5g})".format(self._fP._fX, self._fP._fY, self._fP._fZ, self._fE)

    def __str__(self):
        return repr(self)

    def __eq__(self, other):
        return isinstance(other, Methods) and self.x == other.x and self.y == other.y and self.z == other.z and self.t == other.t

    def _scalar(self, operator, scalar, reverse=False):
        cart = self._to_cartesian()
        if not isinstance(scalar, (numbers.Number, self.awkward0.numpy.number)):
            raise TypeError("cannot {0} a TLorentzVector with a {1}".format(operator.__name__, type(scalar).__name__))
        if reverse:
            return TLorentzVector(operator(scalar, cart.x), operator(scalar, cart.y), operator(scalar, cart.z), operator(scalar, cart.t))
        else:
            return TLorentzVector(operator(cart.x, scalar), operator(cart.y, scalar), operator(cart.z, scalar), operator(cart.t, scalar))

    def _vector(self, operator, vector, reverse=False):
        cart = self._to_cartesian()
        if not isinstance(vector, Methods):
            raise TypeError("cannot {0} a TLorentzVector with a {1}".format(operator.__name__, type(vector).__name__))
        if reverse:
            return TLorentzVector(operator(vector.x, cart.x), operator(vector.y, cart.y), operator(vector.z, cart.z), operator(vector.t, cart.t))
        else:
            return TLorentzVector(operator(cart.x, vector.x), operator(cart.y, vector.y), operator(cart.z, vector.z), operator(cart.t, vector.t))

    def _unary(self, operator):
        cart = self._to_cartesian()
        return TLorentzVector(operator(cart.x), operator(cart.y), operator(cart.z), operator(cart.t))

    @property
    def pt(self):
        return math.sqrt(self.pt2)

    @property
    def eta(self):
        return math.asinh(self.z / math.sqrt(self.x**2 + self.y**2))

    @property
    def phi(self):
        return self.p3.phi

    @property
    def mass(self):
        return math.sqrt(self.mag2)

    @property
    def mag(self):
        return math.sqrt(self.mag2)

    @property
    def mt(self):
        out = self.mt2
        if out >= 0:
            return math.sqrt(out)
        else:
            return -math.sqrt(out)

    @property
    def rapidity(self):
        return math.log((self.t + self.z) / (self.t - self.z)) / 2.0

    @property
    def unit(self):
        return self / math.sqrt(self.mag)

    @property
    def boostp3(self):
        return uproot3_methods.classes.TVector3.TVector3(self.x/self.t, self.y/self.t, self.z/self.t)

    def boost(self, p3):
        if not isinstance(p3, uproot3_methods.classes.TVector3.Methods):
            raise TypeError("boost p3 must be a TVector3")

        b2 = p3.mag2
        gamma = (1.0 - b2)**(-0.5)
        if b2 != 0:
            gamma2 = (gamma - 1.0) / b2
        else:
            gamma2 = 0.0

        bp = self.p3.dot(p3)
        v = self.p3 + gamma2*bp*p3 + gamma*p3*self.t
        return self.__class__(v.x, v.y, v.z, gamma*(self.t + bp))

    @property
    def gamma(self):
        out = self.beta
        if -1 < out < 1:
            return (1 - out**2)**(-0.5)
        else:
            return float("inf")

    def delta_r(self, other):
        return math.sqrt(self.delta_r2(other))

    def rotate_axis(self, axis, angle):
        p3, t = self._rotate_axis(axis, angle)
        x, y, z = p3
        return self.__class__(x, y, z, t)

    def rotate_euler(self, phi=0, theta=0, psi=0):
        p3, t = self._rotate_euler(phi, theta, psi)
        x, y, z = p3
        return self.__class__(x, y, z, t)

    def islightlike(self, tolerance=1e-10):
        return abs(self.mag2) < tolerance

    def __add__(self, other):
        return self._vector(operator.add, other)

    def __radd__(self, other):
        return self._vector(operator.add, other, True)

    def __sub__(self, other):
        return self._vector(operator.sub, other)

    def __rsub__(self, other):
        return self._vector(operator.sub, other, True)

    def __mul__(self, other):
        return self._scalar(operator.mul, other)

    def __rmul__(self, other):
        return self._scalar(operator.mul, other, True)

    def __div__(self, other):
        return self._scalar(operator.div, other)

    def __rdiv__(self, other):
        return self._scalar(operator.div, other, True)

    def __truediv__(self, other):
        return self._scalar(operator.truediv, other)

    def __rtruediv__(self, other):
        return self._scalar(operator.truediv, other, True)

    def __floordiv__(self, other):
        return self._scalar(operator.floordiv, other)

    def __rfloordiv__(self, other):
        return self._scalar(operator.floordiv, other, True)

    def __mod__(self, other):
        return self._scalar(operator.mod, other)

    def __rmod__(self, other):
        return self._scalar(operator.mod, other, True)

    def __divmod__(self, other):
        return self._scalar(operator.divmod, other)

    def __rdivmod__(self, other):
        return self._scalar(operator.divmod, other, True)

    def __pow__(self, other):
        if isinstance(other, (numbers.Number, self.awkward0.numpy.number)):
            if other == 2:
                return self.mag2
            else:
                return self.mag2**(0.5*other)
        else:
            self._scalar(operator.pow, other)

    # no __rpow__

    def __lshift__(self, other):
        return self._scalar(operator.lshift, other)

    def __rlshift__(self, other):
        return self._scalar(operator.lshift, other, True)

    def __rshift__(self, other):
        return self._scalar(operator.rshift, other)

    def __rrshift__(self, other):
        return self._scalar(operator.rshift, other, True)

    def __and__(self, other):
        return self._scalar(operator.and_, other)

    def __rand__(self, other):
        return self._scalar(operator.and_, other, True)

    def __or__(self, other):
        return self._scalar(operator.or_, other)

    def __ror__(self, other):
        return self._scalar(operator.or_, other, True)

    def __xor__(self, other):
        return self._scalar(operator.xor, other)

    def __rxor__(self, other):
        return self._scalar(operator.xor, other, True)

    def __neg__(self):
        return self._unary(operator.neg)

    def __pos__(self):
        return self._unary(operator.pos)

    def __abs__(self):
        return self.mag

    def __invert__(self):
        return self._unary(operator.invert)

class PtEtaPhiMassMethods(Methods):
    _arraymethods = PtEtaPhiMassArrayMethods

    @property
    def pt(self):
        return self._fPt

    @property
    def eta(self):
        return self._fEta

    @property
    def phi(self):
        return self._fPhi

    @property
    def mass(self):
        return self._fMass

    @property
    def mag(self):
        return self._fMass

    @property
    def mag2(self):
        return self._fMass**2

    @property
    def mt(self):
        out = self.mt2
        if out >= 0:
            return math.sqrt(out)
        else:
            return -math.sqrt(out)

    @property
    def mt2(self):
        return self._fMass**2 + self._fPt**2

    @property
    def p3(self):
        return uproot3_methods.classes.TVector3.TVector3(self.x, self.y, self.z)

    @property
    def x(self):
        return self.pt * self.awkward0.numpy.cos(self.phi)

    @property
    def y(self):
        return self.pt * self.awkward0.numpy.sin(self.phi)

    @property
    def z(self):
        return self.pt * self.awkward0.numpy.sinh(self.eta)

    @property
    def t(self):
        x = self.x
        y = self.y
        z = self.z
        mass = self.mass
        return self.awkward0.numpy.sqrt(x*x + y*y + z*z + mass*mass*self.awkward0.numpy.sign(mass))

    def __repr__(self):
        return "PtEtaPhiMassLorentzVector(pt={0:.5g}, eta={1:.5g}, phi={2:.5g}, mass={3:.5g})".format(self._fPt, self._fEta, self._fPhi, self._fMass)

class PtEtaPhiMassLorentzVectorArray(PtEtaPhiMassArrayMethods, uproot3_methods.base.ROOTMethods.awkward0.ObjectArray):
    def __init__(self, pt, eta, phi, mass):
        if isinstance(pt, awkward0.array.jagged.JaggedArray) or isinstance(eta, awkward0.array.jagged.JaggedArray) or isinstance(phi, awkward0.array.jagged.JaggedArray) or isinstance(mass, awkward0.array.jagged.JaggedArray):
            raise TypeError("PtEtaPhiMassLorentzVectorArray constructor arguments must not be jagged; use TLorentzVectorArray.from_ptetaphim for jaggedness-handling")
        self._initObjectArray(self.awkward0.Table())
        self["fPt"]   = pt
        self["fEta"]  = eta
        self["fPhi"]  = phi
        self["fMass"] = mass

    @property
    def pt(self):
        return self["fPt"]

    @pt.setter
    def pt(self, value):
        self["fPt"] = value

    @property
    def eta(self):
        return self["fEta"]

    @eta.setter
    def eta(self, value):
        self["fEta"] = value

    @property
    def phi(self):
        return self["fPhi"]

    @phi.setter
    def phi(self, value):
        self["fPhi"] = value

    @property
    def mass(self):
        return self["fMass"]

    @mass.setter
    def mass(self, value):
        self["fMass"] = value

class TLorentzVectorArray(ArrayMethods, uproot3_methods.base.ROOTMethods.awkward0.ObjectArray):

    def __init__(self, x, y, z, t):
        if isinstance(x, awkward0.array.jagged.JaggedArray) or isinstance(y, awkward0.array.jagged.JaggedArray) or isinstance(z, awkward0.array.jagged.JaggedArray) or isinstance(t, awkward0.array.jagged.JaggedArray):
            raise TypeError("TLorentzVectorArray constructor arguments must not be jagged; use TLorentzVectorArray.from_cartesian for jaggedness-handling")
        self._initObjectArray(self.awkward0.Table())
        self["fX"] = x
        self["fY"] = y
        self["fZ"] = z
        self["fE"] = t

    @classmethod
    def origin(cls, shape, dtype=None):
        if dtype is None:
            dtype = cls.awkward0.numpy.float64
        return cls(cls.awkward0.numpy.zeros(shape, dtype=dtype),
                   cls.awkward0.numpy.zeros(shape, dtype=dtype),
                   cls.awkward0.numpy.zeros(shape, dtype=dtype),
                   cls.awkward0.numpy.zeros(shape, dtype=dtype))

    @classmethod
    def origin_like(cls, array):
        return array * 0.0

    @classmethod
    def from_p3(cls, p3, t):
        return cls.from_cartesian(p3.x, p3.y, p3.z, t)

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_cartesian(cls, x, y, z, t):
        return cls(x, y, z, t)

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_spherical(cls, r, theta, phi, t):
        return cls.from_p3(uproot3_methods.classes.TVector3.TVector3Array.from_spherical(r, theta, phi), t)

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_cylindrical(cls, rho, phi, z, t):
        return cls.from_p3(uproot3_methods.classes.TVector3.TVector3Array.from_cylindrical(rho, phi, z), t)

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_xyzm(cls, x, y, z, m):
        return cls(x, y, z, cls.awkward0.numpy.sqrt(x*x + y*y + z*z + m*m*cls.awkward0.numpy.sign(m)))

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_ptetaphi(cls, pt, eta, phi, energy):
        out = cls(pt * cls.awkward0.numpy.cos(phi),
                  pt * cls.awkward0.numpy.sin(phi),
                  pt * cls.awkward0.numpy.sinh(eta),
                  energy)
        out._memo_pt = pt
        out._memo_eta = eta
        out._memo_phi = phi
        return out

    @classmethod
    def from_ptetaphie(cls, pt, eta, phi, energy):
        return cls.from_ptetaphi(pt, eta, phi, energy)

    @classmethod
    @awkward0.util.wrapjaggedmethod(PtEtaPhiMassJaggedArrayMethods)
    def from_ptetaphim(cls, pt, eta, phi, mass):
        return PtEtaPhiMassLorentzVectorArray(pt,eta,phi,mass)

    @property
    def x(self):
        return self["fX"]

    @x.setter
    def x(self, value):
        self["fX"] = value

    @property
    def y(self):
        return self["fY"]

    @y.setter
    def y(self, value):
        self["fY"] = value

    @property
    def z(self):
        return self["fZ"]

    @z.setter
    def z(self, value):
        self["fZ"] = value

    @property
    def t(self):
        return self["fE"]

    @t.setter
    def t(self, value):
        self["fE"] = value

    @property
    def E(self):
        return self["fE"]

    @E.setter
    def E(self, value):
        self["fE"] = value

class PtEtaPhiMassLorentzVector(PtEtaPhiMassMethods):
    def __init__(self, pt, eta, phi, mass):
        self._fPt   = float(pt)
        self._fEta  = float(eta)
        self._fPhi  = float(phi)
        self._fMass = float(mass)

    @property
    def pt(self):
        return self._fPt

    @pt.setter
    def pt(self,value):
        self._fPt = value

    @property
    def eta(self):
        return self._fEta

    @eta.setter
    def eta(self, value):
        self._fEta = value

    @property
    def phi(self):
        return self._fPhi

    @phi.setter
    def phi(self, value):
        self._fPhi = value

    @property
    def mass(self):
        return self._fMass

    @mass.setter
    def mass(self, value):
        self._fMass = value

class TLorentzVector(Methods):
    def __init__(self, x, y, z, t):
        self._fP = uproot3_methods.classes.TVector3.TVector3(float(x), float(y), float(z))
        self._fE = float(t)

    @classmethod
    def origin(cls):
        return cls(0.0, 0.0, 0.0, 0.0)

    @classmethod
    def from_p3(cls, p3, t):
        out = cls.__new__(cls)
        out._fP = p3
        out._fE = t
        return out

    @classmethod
    def from_spherical(cls, r, theta, phi, t):
        return cls.from_p3(uproot3_methods.classes.TVector3.Methods.from_spherical(r, theta, phi), t)

    @classmethod
    def from_cylindrical(cls, rho, phi, z, t):
        return cls.from_p3(uproot3_methods.classes.TVector3.Methods.from_cylindrical(rho, phi, z), t)

    @classmethod
    def from_xyzm(cls, x, y, z, m):
        return cls(x, y, z, math.sqrt(x*x + y*y + z*z + m*m*(1 if m >= 0 else -1)))

    @classmethod
    def from_ptetaphi(cls, pt, eta, phi, energy):
        return cls(pt * math.cos(phi),
                   pt * math.sin(phi),
                   pt * math.sinh(eta),
                   energy)

    @classmethod
    def from_ptetaphie(cls, pt, eta, phi, energy):
        return cls.from_ptetaphi(pt, eta, phi, energy)

    @classmethod
    def from_ptetaphim(cls, pt, eta, phi, mass):
        return PtEtaPhiMassLorentzVector(pt,eta,phi,mass)

    @property
    def x(self):
        return self._fP._fX

    @x.setter
    def x(self, value):
        self._fP._fX = value

    @property
    def y(self):
        return self._fP._fY

    @y.setter
    def y(self, value):
        self._fP._fY = value

    @property
    def z(self):
        return self._fP._fZ

    @z.setter
    def z(self, value):
        self._fP._fZ = value

    @property
    def t(self):
        return self._fE

    @t.setter
    def t(self, value):
        self._fE = value

    @property
    def E(self):
        return self._fE

    @E.setter
    def E(self, value):
        self._fE = value



import uproot3_methods.base


def _decode(seq):
    if isinstance(seq, bytes):
        return seq.decode("UTF-8")
    else:
        return seq


class Methods(uproot3_methods.base.ROOTMethods):
    def __repr__(self):
        if self._fName is None:
            return "<{0} at 0x{1:012x}>".format(_decode(self._classname), id(self))
        else:
            return "<{0} {1} 0x{2:012x}>".format(
                _decode(self._classname), _decode(self._fName), id(self)
            )

    def __str__(self):
        return str(self._fVal)

    @property
    def name(self):
        return _decode(self._fName)

    @property
    def value(self):
        return self._fVal



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import awkward0.type
import awkward0.array.chunked
import awkward0.array.objects

import uproot3_methods.classes.TLorentzVector

def getcontent(virtual):
    return virtual.array.content

def jaggedtable(rowname, counts, fields):
    Table = counts.Table
    JaggedArray = counts.JaggedArray
    ChunkedArray = counts.ChunkedArray
    VirtualArray = counts.VirtualArray
    VirtualTLorentzVectorArray = awkward0.array.objects.Methods.mixin(uproot3_methods.classes.TLorentzVector.PtEtaPhiMassArrayMethods, VirtualArray)

    countsarray = counts.array
    if isinstance(countsarray, awkward0.array.chunked.ChunkedArray):
        return lazyjagged(countsarray, rowname, [(n, x.array) for n, x in fields])
    else:
        offsets = JaggedArray.counts2offsets(countsarray)
        table = Table.named(rowname)
        for n, x in fields:
            table[n] = VirtualArray(getcontent, x, type=awkward0.type.ArrayType(offsets[-1], x.type.to.to), cache=counts.cache, persistvirtual=counts.persistvirtual)
        columns = table.columns
        if "pt" in columns and "eta" in columns and "phi" in columns and "mass" in columns and "p4" not in columns:
            table["p4"] = VirtualTLorentzVectorArray(uproot3_methods.classes.TLorentzVector.TLorentzVectorArray.from_ptetaphim, (table["pt"], table["eta"], table["phi"], table["mass"]), type=awkward0.type.ArrayType(offsets[-1], uproot3_methods.classes.TLorentzVector.PtEtaPhiMassLorentzVectorArray), cache=counts.cache, persistvirtual=counts.persistvirtual)
        return JaggedArray.fromoffsets(offsets, table)

def lazyjagged(countsarray, rowname, fields):
    ChunkedArray = countsarray.ChunkedArray
    VirtualArray = countsarray.VirtualArray

    chunks = []
    for i, countschunk in enumerate(countsarray.chunks):
        fieldschunks = []
        tabletype = awkward0.type.TableType()
        for fieldname, field in fields:
            assert field.chunksizes[i] == countsarray.chunksizes[i]
            fieldschunks.append((fieldname, field.chunks[i]))
            tabletype[fieldname] = field.type.to.to
        columns = tabletype.columns
        if "pt" in columns and "eta" in columns and "phi" in columns and "mass" in columns and "p4" not in columns:
            tabletype["p4"] = uproot3_methods.classes.TLorentzVector.TLorentzVectorArray.from_ptetaphim
        chunks.append(VirtualArray(jaggedtable, (rowname, countschunk, fieldschunks), type=awkward0.type.ArrayType(len(countschunk), float("inf"), tabletype), cache=countschunk.cache, persistvirtual=countschunk.persistvirtual))
    return ChunkedArray(chunks, countsarray.chunksizes)

def crossref(fromarray, links, subj):
    out = fromarray.array
    ChunkedArray = out.ChunkedArray
    VirtualArray = out.VirtualArray

    if isinstance(out, awkward0.array.chunked.ChunkedArray):
        chunks = []
        for j, chunk in enumerate(out.chunks):
            newtype = awkward0.type.ArrayType(out.chunksizes[j], float("inf"), awkward0.type.TableType())
            for n in chunk.type.to.to.columns:
                newtype.to.to[n] = chunk.type.to.to[n]
            for collection, subname, i, localindex, name, totype in links:
                newtype.to.to[name] = totype
            chunks.append(VirtualArray(crossref, (chunk, links, j), type=newtype, cache=fromarray.cache, persistvirtual=fromarray.persistvirtual))

        return ChunkedArray(chunks, out.chunksizes)

    else:
        for collection, subname, i, localindex, name, totype in links:
            toarray = collection[subname].chunks[i]
            out.content[name] = VirtualArray(indexedmask, (toarray, localindex, subj), type=awkward0.type.ArrayType(out.offsets[-1], totype), cache=fromarray.cache, persistvirtual=fromarray.persistvirtual)
        return out

def indexedmask(toarray, localindex, subj):
    jagged = toarray.array
    localindex = localindex.array
    if subj is not None:
        jagged = jagged.chunks[subj].array
        localindex = localindex.chunks[subj].array

    globalindex = localindex + jagged.starts
    globalindex.content[localindex.content < 0] = -1
    return toarray.IndexedMaskedArray(globalindex.content, jagged.content)

def transform(array):
    array._valid()
    array.check_whole_valid = False

    Table = array.Table
    VirtualArray = array.VirtualArray

    stuff = [("run",                               "run",                                None),
             ("luminosityBlock",                   "lumi",                               None),
             ("event",                             "event",                              None),
             ("Electron_",                         "electrons",                          []),
             ("Muon_",                             "muons",                              []),
             ("Tau_",                              "taus",                               []),
             ("Photon_",                           "photons",                            []),
             ("Jet_",                              "jets",                               []),
             ("FatJet_",                           "fatjets",                            []),
             ("SubJet_",                           "subjets",                            []),
             ("IsoTrack_",                         "isotracks",                          []),
             ("SoftActivityJet_",                  "softjets",                           []),
             ("SoftActivityJetHT",                 "softactivity.HT",                    None),
             ("SoftActivityJetHT2",                "softactivity.HT2",                   None),
             ("SoftActivityJetHT5",                "softactivity.HT5",                   None),
             ("SoftActivityJetHT10",               "softactivity.HT10",                  None),
             ("SoftActivityJetNjets2",             "softactivity.njets2",                None),
             ("SoftActivityJetNjets5",             "softactivity.njets5",                None),
             ("SoftActivityJetNjets10",            "softactivity.njets10",               None),
             ("fixedGridRhoFastjetAll",            "fixedGridRhoFastjet.everything",     None),
             ("fixedGridRhoFastjetCentralCalo",    "fixedGridRhoFastjet.centralcalo",    None),
             ("fixedGridRhoFastjetCentralNeutral", "fixedGridRhoFastjet.centralneutral", None),
             ("MET_",                              "MET",                                Table.named("MET")),
             ("RawMET_",                           "rawMET",                             Table.named("RawMET")),
             ("CaloMET_",                          "caloMET",                            Table.named("CaloMET")),
             ("PuppiMET_",                         "puppiMET",                           Table.named("PuppiMET")),
             ("TkMET_",                            "tkMET",                              Table.named("TkMET")),
             ("PV_",                               "PV",                                 Table.named("PV")),
             ("SV_",                               "SVs",                                []),
             ("OtherPV_",                          "otherPVs",                           []),
             ("Pileup_",                           "pileup",                             Table.named("Pileup")),
             ("Flag_",                             "flags",                              Table.named("Flags")),
             ("TrigObj_",                          "trigobjs",                           []),
             ("HLT_",                              "HLT",                                Table.named("HLT")),
             ("HLTriggerFirstPath",                "HLT.firstpath",                      None),
             ("HLTriggerFinalPath",                "HLT.finalpath",                      None),
             ("Generator_",                        "gen",                                Table.named("Generator")),
             ("GenDressedLepton_",                 "gen.dressedleptons",                 []),
             ("GenPart_",                          "gen.partons",                        []),
             ("GenJet_",                           "gen.jets",                           []),
             ("GenJetAK8_",                        "gen.jetsAK8",                        []),
             ("SubGenJetAK8_",                     "gen.subjetsAK8",                     []),
             ("GenVisTau_",                        "gen.vistaus",                        []),
             ("GenMET_",                           "gen.MET",                            Table.named("GenMET")),
             ("LHE_",                              "gen.LHE",                            Table.named("LHE")),
             ("LHEPart_",                          "gen.LHEpartons",                     []),
             ("genWeight",                         "gen.genweight",                      None),
             ("LHEPdfWeight",                      "gen.LHEpdfweight",                   None),
             ("LHEScaleWeight",                    "gen.LHEscaleweight",                 None),
             ("LHEWeight_originalXWGTUP",          "gen.LHEweight_originalXWGTUP",       None),
             ]

    others = []
    for n in array.columns:
        for prefix, rename, data in stuff:
            if n.startswith(prefix):
                if data is None:
                    pass
                elif isinstance(data, list):
                    data.append((n[len(prefix):], array[n]))
                else:
                    data[n[len(prefix):]] = array[n]
                break
            elif n == "n" + prefix.rstrip("_"):
                break
        else:
            others.append(n)

    events = Table.named("Event")

    def makecollection(rename):
        if "." in rename:
            outer, inner = rename.split(".")
            if outer not in events.columns:
                events[outer] = Table.named(outer.capitalize())
            return events[outer], inner
        else:
            return events, rename

    for prefix, rename, data in stuff:
        if data is None:
            if prefix in array.columns:
                collection, rename = makecollection(rename)
                collection[rename] = array[prefix]
        elif isinstance(data, list):
            rowname = prefix[:-1]
            countname = "n" + rowname
            if len(data) > 0 and countname in array.columns:
                collection, rename = makecollection(rename)
                collection[rename] = lazyjagged(array[countname], rowname, data)
        else:
            if len(data.columns) > 0:
                collection, rename = makecollection(rename)
                collection[rename] = data

    eventtype = events.type

    eventtype.to["electrons"].to["photon"] = awkward0.type.OptionType(eventtype.to["photons"].to)
    eventtype.to["electrons"].to["photon"].check = False
    eventtype.to["electrons"].to["jet"] = awkward0.type.OptionType(eventtype.to["jets"].to)
    eventtype.to["electrons"].to["jet"].check = False
    for i, chunk in enumerate(events["electrons"].chunks):
        assert events["electrons"].chunksizes[i] == events["jets"].chunksizes[i] == events["photons"].chunksizes[i]
        events["electrons"].chunks[i] = VirtualArray(crossref, (chunk, [
            (events, "photons", i, array["Electron_photonIdx"].chunks[i], "photon", eventtype.to["electrons"].to["photon"]),
            (events, "jets", i, array["Electron_jetIdx"].chunks[i], "jet", eventtype.to["electrons"].to["jet"]),
            ], None), type=awkward0.type.ArrayType(events["electrons"].chunksizes[i], eventtype.to["electrons"]), cache=chunk.cache, persistvirtual=chunk.persistvirtual)

    eventtype.to["muons"].to["jet"] = awkward0.type.OptionType(eventtype.to["jets"].to)
    eventtype.to["muons"].to["jet"].check = False
    for i, chunk in enumerate(events["muons"].chunks):
        assert events["muons"].chunksizes[i] == events["jets"].chunksizes[i]
        events["muons"].chunks[i] = VirtualArray(crossref, (chunk, [
            (events, "jets", i, array["Muon_jetIdx"].chunks[i], "jet", eventtype.to["muons"].to["jet"]),
            ], None), type=awkward0.type.ArrayType(events["muons"].chunksizes[i], eventtype.to["muons"]), cache=chunk.cache, persistvirtual=chunk.persistvirtual)

    eventtype.to["taus"].to["jet"] = awkward0.type.OptionType(eventtype.to["jets"].to)
    eventtype.to["taus"].to["jet"].check = False
    for i, chunk in enumerate(events["taus"].chunks):
        assert events["taus"].chunksizes[i] == events["jets"].chunksizes[i]
        events["taus"].chunks[i] = VirtualArray(crossref, (chunk, [
            (events, "jets", i, array["Tau_jetIdx"].chunks[i], "jet", eventtype.to["taus"].to["jet"]),
            ], None), type=awkward0.type.ArrayType(events["jets"].chunksizes[i], eventtype.to["taus"]), cache=chunk.cache, persistvirtual=chunk.persistvirtual)

    eventtype.to["taus"].to["jet"] = awkward0.type.OptionType(eventtype.to["jets"].to)
    eventtype.to["taus"].to["jet"].check = False
    for i, chunk in enumerate(events["taus"].chunks):
        assert events["taus"].chunksizes[i] == events["jets"].chunksizes[i]
        events["taus"].chunks[i] = VirtualArray(crossref, (chunk, [
            (events, "jets", i, array["Tau_jetIdx"].chunks[i], "jet", eventtype.to["taus"].to["jet"]),
            ], None), type=awkward0.type.ArrayType(events["taus"].chunksizes[i], eventtype.to["taus"]), cache=chunk.cache, persistvirtual=chunk.persistvirtual)

    eventtype.to["photons"].to["electron"] = awkward0.type.OptionType(eventtype.to["electrons"].to)
    eventtype.to["photons"].to["electron"].check = False
    eventtype.to["photons"].to["jet"] = awkward0.type.OptionType(eventtype.to["jets"].to)
    eventtype.to["photons"].to["jet"].check = False
    for i, chunk in enumerate(events["photons"].chunks):
        assert events["photons"].chunksizes[i] == events["jets"].chunksizes[i] == events["electrons"].chunksizes[i]
        events["photons"].chunks[i] = VirtualArray(crossref, (chunk, [
            (events, "electrons", i, array["Photon_electronIdx"].chunks[i], "electron", eventtype.to["photons"].to["electron"]),
            (events, "jets", i, array["Photon_jetIdx"].chunks[i], "jet", eventtype.to["photons"].to["jet"]),
            ], None), type=awkward0.type.ArrayType(events["photons"].chunksizes[i], eventtype.to["photons"]), cache=chunk.cache, persistvirtual=chunk.persistvirtual)

    eventtype.to["jets"].to["electron1"] = awkward0.type.OptionType(eventtype.to["electrons"].to)
    eventtype.to["jets"].to["electron1"].check = False
    eventtype.to["jets"].to["electron2"] = awkward0.type.OptionType(eventtype.to["electrons"].to)
    eventtype.to["jets"].to["electron2"].check = False
    eventtype.to["jets"].to["muon1"] = awkward0.type.OptionType(eventtype.to["muons"].to)
    eventtype.to["jets"].to["muon1"].check = False
    eventtype.to["jets"].to["muon2"] = awkward0.type.OptionType(eventtype.to["muons"].to)
    eventtype.to["jets"].to["muon2"].check = False
    for i, chunk in enumerate(events["jets"].chunks):
        assert events["jets"].chunksizes[i] == events["electrons"].chunksizes[i] == events["muons"].chunksizes[i]
        events["jets"].chunks[i] = VirtualArray(crossref, (chunk, [
            (events, "electrons", i, array["Jet_electronIdx1"].chunks[i], "electron1", eventtype.to["jets"].to["electron1"]),
            (events, "electrons", i, array["Jet_electronIdx2"].chunks[i], "electron2", eventtype.to["jets"].to["electron2"]),
            (events, "muons", i, array["Jet_muonIdx1"].chunks[i], "muon1", eventtype.to["jets"].to["muon1"]),
            (events, "muons", i, array["Jet_muonIdx2"].chunks[i], "muon2", eventtype.to["jets"].to["muon2"]),
            ], None), type=awkward0.type.ArrayType(events["jets"].chunksizes[i], eventtype.to["jets"]), cache=chunk.cache, persistvirtual=chunk.persistvirtual)

    eventtype.to["fatjets"].to["subjet1"] = awkward0.type.OptionType(eventtype.to["jets"].to)
    eventtype.to["fatjets"].to["subjet1"].check = False
    eventtype.to["fatjets"].to["subjet2"] = awkward0.type.OptionType(eventtype.to["jets"].to)
    eventtype.to["fatjets"].to["subjet2"].check = False
    for i, chunk in enumerate(events["fatjets"].chunks):
        assert events["fatjets"].chunksizes[i] == events["jets"].chunksizes[i]
        events["fatjets"].chunks[i] = VirtualArray(crossref, (chunk, [
            (events, "jets", i, array["FatJet_subJetIdx1"].chunks[i], "subjet1", eventtype.to["fatjets"].to["subjet1"]),
            (events, "jets", i, array["FatJet_subJetIdx2"].chunks[i], "subjet2", eventtype.to["fatjets"].to["subjet2"]),
            ], None), type=awkward0.type.ArrayType(events["fatjets"].chunksizes[i], eventtype.to["fatjets"]), cache=chunk.cache, persistvirtual=chunk.persistvirtual)

    if len(others) > 0:
        etc = events["etc"] = Table.named("OtherFields")
        for n in others:
            etc[n] = array[n]
    events["raw"] = array

    return events



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import importlib
import re

def transformer(name):
    m = re.match(r"^([a-zA-Z_][a-zA-Z_0-9]*)(\.[a-zA-Z_][a-zA-Z_0-9]*)*$", name)
    if m is None:
        raise ValueError("profile name must match \"identifier(.identifier)*\"")
    return getattr(importlib.import_module("uproot3_methods.profiles." + m.string), "transform")



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

def hasmethods(name):
    if name not in globals():
        if name in hasmethods.loaders:
            globals()[name] = hasmethods.loaders[name].load_module(name)
        elif '_3c_' in name and '_3e_' in name:
            bare_name = name.split('_3c_')[0]
            if bare_name in hasmethods.loaders:
                globals()[name] = hasmethods.loaders[bare_name].load_module(bare_name)

    return name in globals() and isinstance(getattr(globals()[name], "Methods", None), type)

import pkgutil

hasmethods.loaders = dict([(module_name, loader.find_module(module_name)) for loader, module_name, is_pkg in pkgutil.walk_packages(__path__)])

del pkgutil



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import uproot3_methods.base

class Methods(uproot3_methods.base.ROOTMethods):

	@property
	def xerrorshigh(self):
		return self._fEXhigh

	@property
	def xerrorslow(self):
		return self._fEXlow

	@property
	def yerrorshigh(self):
		return self._fEYhigh

	@property
	def yerrorslow(self):
		return self._fEYlow

	def matplotlib(self, showtitle=True, show=False, **kwargs):
		import matplotlib.pyplot as pyplot

		_xerrs = [self.xerrorslow, self.xerrorshigh]
		_yerrs = [self.yerrorslow, self.yerrorshigh]

		_xlabel = _decode(self.xlabel if self.xlabel is not None else "")
		_ylabel = _decode(self.ylabel if self.ylabel is not None else "")

		pyplot.errorbar(self.xvalues, self.yvalues, xerr=_xerrs, yerr=_yerrs, **kwargs)
		pyplot.xlabel(_xlabel)
		pyplot.ylabel(_ylabel)
		if showtitle:
			_title = _decode(self.title)
			pyplot.title(_title)

		if show:
			pyplot.show()

def _decode(sequence):
	return sequence.decode() if isinstance(sequence, bytes) else sequence



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import numpy

import uproot3_methods.base

class Methods(uproot3_methods.base.ROOTMethods):
    def hello(self):
        return "world", len(dir(self))



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import math
import numbers
import operator

class Common(object):
    @property
    def mag2(self):
        return self.dot(self)

    @property
    def mag(self):
        return self.awkward0.numpy.sqrt(self.mag2)

    @property
    def rho2(self):
        out = self.x*self.x
        out = out + self.y*self.y
        return out

    def delta_phi(self, other):
        return (self.phi - other.phi + math.pi) % (2*math.pi) - math.pi

    def isparallel(self, other, tolerance=1e-10):
        return 1 - self.cosdelta(other) < tolerance

    def isantiparallel(self, other, tolerance=1e-10):
        return self.cosdelta(other) - (-1) < tolerance

    def iscollinear(self, other, tolerance=1e-10):
        return 1 - self.awkward0.numpy.absolute(self.cosdelta(other)) < tolerance

    def __lt__(self, other):
        raise TypeError("spatial vectors have no natural ordering")

    def __gt__(self, other):
        raise TypeError("spatial vectors have no natural ordering")

    def __le__(self, other):
        raise TypeError("spatial vectors have no natural ordering")

    def __ge__(self, other):
        raise TypeError("spatial vectors have no natural ordering")

class ArrayMethods(Common):
    @property
    def unit(self):
        return self / self.mag

    @property
    def rho(self):
        out = self.rho2
        return self.awkward0.numpy.sqrt(out)

    @property
    def phi(self):
        return self.awkward0.numpy.arctan2(self.y, self.x)

    def cosdelta(self, other):
        denom = self.mag2 * other.mag2
        mask = (denom > 0)
        denom = denom[mask]
        denom[:] = self.awkward0.numpy.sqrt(denom)

        out = self.dot(other)
        out[mask] /= denom

        mask = self.awkward0.numpy.logical_not(mask)
        out[mask] = 1

        return self.awkward0.numpy.clip(out, -1, 1)

    def angle(self, other, normal=None, degrees=False):
        out = self.awkward0.numpy.arccos(self.cosdelta(other))
        if normal is not None:
            a = self.unit
            b = other.unit
            out = out * self.awkward0.numpy.sign(normal.dot(a.cross(b)))
        if degrees:
            out = self.awkward0.numpy.multiply(out, 180.0/self.awkward0.numpy.pi)
        return out

    def isopposite(self, other, tolerance=1e-10):
        tmp = self + other
        tmp.x = self.awkward0.numpy.absolute(tmp.x)
        tmp.y = self.awkward0.numpy.absolute(tmp.y)
        tmp.z = self.awkward0.numpy.absolute(tmp.z)

        out = (tmp.x < tolerance)
        out = self.awkward0.numpy.bitwise_and(out, tmp.y < tolerance)
        out = self.awkward0.numpy.bitwise_and(out, tmp.z < tolerance)
        return out

    def isperpendicular(self, other, tolerance=1e-10):
        tmp = self.dot(other)
        tmp.x = self.awkward0.numpy.absolute(tmp.x)
        tmp.y = self.awkward0.numpy.absolute(tmp.y)
        tmp.z = self.awkward0.numpy.absolute(tmp.z)

        out = (tmp.x < tolerance)
        out = self.awkward0.numpy.bitwise_and(out, tmp.y < tolerance)
        out = self.awkward0.numpy.bitwise_and(out, tmp.z < tolerance)
        return out

class Methods(Common):
    @property
    def unit(self):
        return self / self.mag

    @property
    def rho(self):
        return math.sqrt(self.rho2)

    @property
    def phi(self):
        return math.atan2(self.y, self.x)

    def cosdelta(self, other):
        m1 = self.mag2
        m2 = other.mag2
        if m1 == 0 or m2 == 0:
            return 1.0
        r = self.dot(other) / math.sqrt(m1 * m2)
        return max(-1.0, min(1.0, r))

    def angle(self, other, degrees=False):
        out = math.acos(self.cosdelta(other))
        if degrees:
            out = out * 180.0/math.pi
        return out

    def isopposite(self, other, tolerance=1e-10):
        tmp = self + other
        return abs(tmp.x) < tolerance and abs(tmp.y) < tolerance and abs(tmp.z) < tolerance

    def isperpendicular(self, other, tolerance=1e-10):
        tmp = self.dot(other)
        return abs(tmp.x) < tolerance and abs(tmp.y) < tolerance and abs(tmp.z) < tolerance

    def __add__(self, other):
        return self._vector(operator.add, other)

    def __radd__(self, other):
        return self._vector(operator.add, other, True)

    def __sub__(self, other):
        return self._vector(operator.sub, other)

    def __rsub__(self, other):
        return self._vector(operator.sub, other, True)

    def __mul__(self, other):
        return self._scalar(operator.mul, other)

    def __rmul__(self, other):
        return self._scalar(operator.mul, other, True)

    def __div__(self, other):
        return self._scalar(operator.div, other)

    def __rdiv__(self, other):
        return self._scalar(operator.div, other, True)

    def __truediv__(self, other):
        return self._scalar(operator.truediv, other)

    def __rtruediv__(self, other):
        return self._scalar(operator.truediv, other, True)

    def __floordiv__(self, other):
        return self._scalar(operator.floordiv, other)

    def __rfloordiv__(self, other):
        return self._scalar(operator.floordiv, other, True)

    def __mod__(self, other):
        return self._scalar(operator.mod, other)

    def __rmod__(self, other):
        return self._scalar(operator.mod, other, True)

    def __divmod__(self, other):
        return self._scalar(operator.divmod, other)

    def __rdivmod__(self, other):
        return self._scalar(operator.divmod, other, True)

    def __pow__(self, other):
        if isinstance(other, (numbers.Number, self.awkward0.numpy.number)):
            if other == 2:
                return self.mag2
            else:
                return self.mag2**(0.5*other)
        else:
            self._scalar(operator.pow, other)

    # no __rpow__

    def __lshift__(self, other):
        return self._scalar(operator.lshift, other)

    def __rlshift__(self, other):
        return self._scalar(operator.lshift, other, True)

    def __rshift__(self, other):
        return self._scalar(operator.rshift, other)

    def __rrshift__(self, other):
        return self._scalar(operator.rshift, other, True)

    def __and__(self, other):
        return self._scalar(operator.and_, other)

    def __rand__(self, other):
        return self._scalar(operator.and_, other, True)

    def __or__(self, other):
        return self._scalar(operator.or_, other)

    def __ror__(self, other):
        return self._scalar(operator.or_, other, True)

    def __xor__(self, other):
        return self._scalar(operator.xor, other)

    def __rxor__(self, other):
        return self._scalar(operator.xor, other, True)

    def __neg__(self):
        return self._unary(operator.neg)

    def __pos__(self):
        return self._unary(operator.pos)

    def __abs__(self):
        return self.mag

    def __invert__(self):
        return self._unary(operator.invert)



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import numpy

import uproot3_methods.base


class Methods(uproot3_methods.base.ROOTMethods):
    @property
    def numbins(self):
        return self.xnumbins * self.ynumbins * self.znumbins

    @property
    def xnumbins(self):
        return self._fXaxis._fNbins

    @property
    def ynumbins(self):
        return self._fYaxis._fNbins

    @property
    def znumbins(self):
        return self._fZaxis._fNbins

    @property
    def low(self):
        return self.xlow, self.ylow, self.zlow

    @property
    def high(self):
        return self.xhigh, self.yhigh, self.zlow

    @property
    def xlow(self):
        return self._fXaxis._fXmin

    @property
    def xhigh(self):
        return self._fXaxis._fXmax

    @property
    def ylow(self):
        return self._fYaxis._fXmin

    @property
    def yhigh(self):
        return self._fYaxis._fXmax

    @property
    def zlow(self):
        return self._fZaxis._fXmin

    @property
    def zhigh(self):
        return self._fZaxis._fXmax

    @property
    def underflows(self):
        uf = numpy.array(self.allvalues)
        xuf = uf[0]
        yuf = uf[:, 0, :]
        zuf = uf[:, :, 0]
        return xuf, yuf, zuf

    @property
    def xunderflows(self):
        return self.underflows[0]

    @property
    def yunderflows(self):
        return self.underflows[1]

    @property
    def zunderflows(self):
        return self.underflows[2]

    @property
    def overflows(self):
        of = numpy.array(self.allvalues)
        xof = of[-1]
        yof = of[:, -1, :]
        zof = of[:, :, -1]
        return xof, yof, zof

    @property
    def xoverflows(self):
        return self.overflows[0]

    @property
    def yoverflows(self):
        return self.overflows[1]

    @property
    def zoverflows(self):
        return self.overflows[2]

    @property
    def edges(self):
        xaxis = self._fXaxis
        yaxis = self._fYaxis
        zaxis = self._fZaxis
        if len(getattr(xaxis, "_fXbins", [])) > 0:
            xedges = numpy.array(xaxis._fXbins)
        else:
            xedges = numpy.linspace(xaxis._fXmin, xaxis._fXmax,
                                    xaxis._fNbins + 1)

        if len(getattr(yaxis, "_fXbins", [])) > 0:
            yedges = numpy.array(yaxis._fXbins)
        else:
            yedges = numpy.linspace(yaxis._fXmin, yaxis._fXmax,
                                    yaxis._fNbins + 1)

        if len(getattr(zaxis, "_fXbins", [])) > 0:
            zedges = numpy.array(zaxis._fXbins)
        else:
            zedges = numpy.linspace(zaxis._fXmin, zaxis._fXmax,
                                    zaxis._fNbins + 1)

        return xedges, yedges, zedges

    @property
    def alledges(self):
        xedges, yedges, zedges = self.edges
        vx = numpy.empty(len(xedges) + 2)
        vy = numpy.empty(len(yedges) + 2)
        vz = numpy.empty(len(zedges) + 2)
        vx[0] = -numpy.inf
        vx[-1] = numpy.inf
        vx[1:-1] = xedges
        vy[0] = -numpy.inf
        vy[-1] = numpy.inf
        vy[1:-1] = yedges
        vz[0] = -numpy.inf
        vz[-1] = numpy.inf
        vz[1:-1] = zedges
        return vx, vy, vz

    @property
    def bins(self):
        xedges, yedges, zedges = self.edges
        vx = numpy.empty((len(xedges) - 1, 2))
        vy = numpy.empty((len(yedges) - 1, 2))
        vz = numpy.empty((len(zedges) - 1, 2))
        vx[:, 0] = xedges[:-1]
        vx[:, 1] = xedges[1:]
        vy[:, 0] = yedges[:-1]
        vy[:, 1] = yedges[1:]
        vz[:, 0] = zedges[:-1]
        vz[:, 1] = zedges[1:]
        return vx, vy, vz

    @property
    def allbins(self):
        xedges, yedges, zedges = self.alledges
        vx = numpy.empty((len(xedges) - 1, 2))
        vy = numpy.empty((len(yedges) - 1, 2))
        vz = numpy.empty((len(zedges) - 1, 2))
        vx[:, 0] = xedges[:-1]
        vx[:, 1] = xedges[1:]
        vy[:, 0] = yedges[:-1]
        vy[:, 1] = yedges[1:]
        vz[:, 0] = zedges[:-1]
        vz[:, 1] = zedges[1:]
        return vx, vy, vz

    @property
    def values(self):
        va = self.allvalues
        return va[1:self.xnumbins+1, 1:self.ynumbins+1, 1:self.znumbins+1]

    @property
    def allvalues(self):
        dtype = getattr(self, "_dtype", numpy.dtype(numpy.float64))
        v = numpy.array(self[:], dtype=dtype.newbyteorder("="))
        v = v.reshape(self.znumbins + 2, self.ynumbins + 2, self.xnumbins + 2)
        return v.T

    @property
    def variances(self):
        va = self.allvariances
        return va[1:self.xnumbins+1, 1:self.ynumbins+1, 1:self.znumbins+1]

    @property
    def allvariances(self):
        if len(getattr(self, "_fSumw2", [])) != len(self):
            v = numpy.array(self, dtype=numpy.float64)
        else:
            v = numpy.array(self._fSumw2, dtype=numpy.float64)
        v = v.reshape(self.znumbins + 2, self.ynumbins + 2, self.xnumbins + 2)
        return v.T

    def numpy(self):
        return (self.values, [self.edges])

    def allnumpy(self):
        return (self.allvalues, [self.alledges])

    def interval(self, index, axis):
        if axis == "x":
            low = self.xlow
            high = self.xhigh
            nbins = self.xnumbins
            bins = self._fXaxis._fXbins
        elif axis == "y":
            low = self.ylow
            high = self.yhigh
            nbins = self.ynumbins
            bins = self._fYaxis._fXbins
        elif axis == "z":
            low = self.zlow
            high = self.zhigh
            nbins = self.znumbins
            bins = self._fZaxis._fXbins
        else:
            raise ValueError("axis must be 'x','y' or 'z'")

        if index < 0:
            index += nbins

        if index == 0:
            return float("-inf"), low
        elif index == nbins + 1:
            return high, float("inf")
        else:
            if not bins:
                norm = float(high-low) / nbins
                xedges = (index-1)*norm + low, index*norm + low
            else:
                xedges = bins[index-1], bins[index]
            return xedges

    def xinterval(self, index):
        return self.interval(index, "x")

    def yinterval(self, index):
        return self.interval(index, "y")

    def zinterval(self, index):
        return self.interval(index, "z")

    @property
    def ylabels(self):
        if self._fYaxis._fLabels is None:
            return None
        else:
            return [str(x) for x in self._fYaxis._fLabels]

    @property
    def zlabels(self):
        if self._fZaxis._fLabels is None:
            return None
        else:
            return [str(x) for x in self._fZaxis._fLabels]



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import uproot3_methods.base

class Methods(uproot3_methods.base.ROOTMethods):
	def __repr__(self):
		if self._fName is None:
			return "<{0} at 0x{1:012x}>".format(self._classname, id(self))
		else:
			return "<{0} {1} 0x{2:012x}>".format(self._classname, repr(self._fName), id(self))

	@property
	def name(self):
		return self._fName

	@property
	def title(self):
		return self._fTitle

	@property
	def maximum(self):
		return self._fMaximum

	@property
	def minimum(self):
		return self._fMinimum

	@property
	def npoints(self):
		return self._fNpoints

	@property
	def xvalues(self):
		return self._fX

	@property
	def yvalues(self):
		return self._fY

	@property
	def xlabel(self):
		if self._fHistogram is None:
			return None
		elif getattr(self._fHistogram, "_fXaxis", None) is None:
			return None
		else:
			return getattr(self._fHistogram._fXaxis, "_fTitle", None)

	@property
	def ylabel(self):
		if self._fHistogram is None:
			return None
		elif getattr(self._fHistogram, "_fYaxis", None) is None:
			return None
		else:
			return getattr(self._fHistogram._fYaxis, "_fTitle", None)

	def matplotlib(self, showtitle=True, show=False, fmt="", **kwargs):
		import matplotlib.pyplot as pyplot

		_xlabel = _decode(self.xlabel if self.xlabel is not None else "")
		_ylabel = _decode(self.ylabel if self.ylabel is not None else "")

		pyplot.plot(self.xvalues, self.yvalues, fmt, **kwargs)
		pyplot.xlabel(_xlabel)
		pyplot.ylabel(_ylabel)
		if showtitle:
			_title = _decode(self.title)
			pyplot.title(_title)

		if show:
			pyplot.show()

def _decode(sequence):
	return sequence.decode() if isinstance(sequence, bytes) else sequence



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import importlib

# uses numpy, not awkward0.numpy, because it operates on TFile data, not TTree data
import numpy

def towriteable(obj):
    def identity(x):
        return x

    def resolve(obj):
        def types(cls, obj):
            if cls is numpy.ndarray:
                yield ("numpy", "ndarray", len(obj.shape), str(obj.dtype))
            elif cls.__module__ == "pandas.core.frame" and cls.__name__ == "DataFrame":
                yield ("pandas.core.frame", "DataFrame", obj.index.__class__.__name__, set(obj.columns))
            else:
                yield (cls.__module__, cls.__name__)
            for x in cls.__bases__:
                for y in types(x, obj):
                    yield y

        if any(x == ("builtins", "bytes") or x == ("builtins", "str") or x == ("__builtin__", "str") or x == ("__builtin__", "unicode") for x in types(obj.__class__, obj)):
            return (None, None, "uproot3.write.objects.TObjString", "TObjString")

        # made with numpy.histogram
        elif isinstance(obj, tuple) and len(obj) == 2 and any(x[:2] == ("numpy", "ndarray") for x in types(obj[0].__class__, obj[0])) and any(x[:2] == ("numpy", "ndarray") for x in types(obj[1].__class__, obj[1])) and len(obj[0].shape) == 1 and len(obj[1].shape) == 1 and obj[0].shape[0] == obj[1].shape[0] - 1:
            return ("uproot3_methods.classes.TH1", "from_numpy", "uproot3.write.objects.TH", "TH")

        # made with numpy.histogram2d
        elif isinstance(obj, tuple) and len(obj) == 3 and any(x[:2] == ("numpy", "ndarray") for x in types(obj[0].__class__, obj[0])) and any(x[:2] == ("numpy", "ndarray") for x in types(obj[1].__class__, obj[1])) and any(x[:2] == ("numpy", "ndarray") for x in types(obj[2].__class__, obj[2])) and len(obj[0].shape) == 2 and len(obj[1].shape) == 1 and len(obj[2].shape) == 1 and obj[0].shape[0] == obj[1].shape[0] - 1 and obj[0].shape[1] == obj[2].shape[0] - 1:
            return ("uproot3_methods.classes.TH2", "from_numpy", "uproot3.write.objects.TH", "TH")

        # made with numpy.histogramdd (2-dimensional)
        elif isinstance(obj, tuple) and len(obj) == 2 and any(x[:2] == ("numpy", "ndarray") for x in types(obj[0].__class__, obj[0])) and isinstance(obj[1], list) and len(obj[1]) == 2 and any(x[:2] == ("numpy", "ndarray") for x in types(obj[1][0].__class__, obj[1][0])) and any(x[:2] == ("numpy", "ndarray") for x in types(obj[1][1].__class__, obj[1][1])) and len(obj[0].shape) == 2 and len(obj[1][0].shape) == 1 and len(obj[1][1].shape) == 1 and obj[0].shape[0] == obj[1][0].shape[0] - 1 and obj[0].shape[1] == obj[1][1].shape[0] - 1:
            return ("uproot3_methods.classes.TH2", "from_numpy", "uproot3.write.objects.TH", "TH")

        elif any(x[:3] == ("pandas.core.frame", "DataFrame", "IntervalIndex") and "count" in x[3] for x in types(obj.__class__, obj)):
            return ("uproot3_methods.classes.TH1", "from_pandas", "uproot3.write.objects.TH", "TH")

        elif any(x == ("physt.histogram1d", "Histogram1D") for x in types(obj.__class__, obj)):
            return ("uproot3_methods.classes.TH1", "from_physt", "uproot3.write.objects.TH", "TH")

        elif any(x == ("uproot3_methods.classes.TH1", "Methods") or x == ("TH1", "Methods") for x in types(obj.__class__, obj)):
            return (None, None, "uproot3.write.objects.TH", "TH")

        elif any(x == ("uproot3_methods.classes.TH2", "Methods") or x == ("TH2", "Methods") for x in types(obj.__class__, obj)):
            return (None, None, "uproot3.write.objects.TH", "TH")

        elif any(x == ("uproot3_methods.classes.TH3", "Methods") or x == ("TH3", "Methods") for x in types(obj.__class__, obj)):
            return (None, None, "uproot3.write.objects.TH", "TH")

        else:
            raise TypeError("type {0} from module {1} is not writeable by uproot3".format(obj.__class__.__name__, obj.__class__.__module__))

    convertmod, convertfcn, mod, cls = resolve(obj)

    if convertfcn is not None:
        convert = getattr(importlib.import_module(convertmod), convertfcn)
        obj = convert(obj)

    cls = getattr(importlib.import_module(mod), cls)
    return cls(obj)



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import numbers
import math
import sys

import numpy

import uproot3_methods.base

class Methods(uproot3_methods.base.ROOTMethods):
    def __repr__(self):
        if self.name is None:
            return "<{0} at 0x{1:012x}>".format(self._classname, id(self))
        else:
            return "<{0} {1} 0x{2:012x}>".format(self._classname, repr(self._fName), id(self))

    @property
    def name(self):
        return getattr(self, "_fName", None)

    @property
    def title(self):
        return self._fTitle

    @property
    def numbins(self):
        return self._fXaxis._fNbins

    @property
    def low(self):
        return self._fXaxis._fXmin

    @property
    def high(self):
        return self._fXaxis._fXmax

    @property
    def underflows(self):
        return self[0]

    @property
    def overflows(self):
        return self[-1]

    @property
    def edges(self):
        axis = self._fXaxis
        if len(getattr(axis, "_fXbins", [])) > 0:
            return numpy.array(axis._fXbins)
        else:
            return numpy.linspace(axis._fXmin, axis._fXmax, axis._fNbins + 1)

    @property
    def alledges(self):
        axis = self._fXaxis
        v = numpy.empty(axis._fNbins + 3)
        v[0] = -numpy.inf
        v[-1] = numpy.inf
        v[1:-1] = self.edges
        return v

    @property
    def bins(self):
        edges = self.edges
        out = numpy.empty((len(edges) - 1, 2))
        out[:, 0] = edges[:-1]
        out[:, 1] = edges[1:]
        return out

    @property
    def allbins(self):
        edges = self.alledges
        out = numpy.empty((len(edges) - 1, 2))
        out[:, 0] = edges[:-1]
        out[:, 1] = edges[1:]
        return out

    @property
    def values(self):
        return self.allvalues[1:-1]

    @property
    def allvalues(self):
        return numpy.array(self, dtype=getattr(self, "_dtype", numpy.dtype(numpy.float64)).newbyteorder("="))

    @property
    def variances(self):
        return self.allvariances[1:-1]

    @property
    def allvariances(self):
        if len(getattr(self, "_fSumw2", [])) != len(self):
            return numpy.array(self, dtype=numpy.float64)
        else:
            return numpy.array(self._fSumw2, dtype=numpy.float64)

    def numpy(self):
        return self.values, self.edges

    def allnumpy(self):
        return self.allvalues, self.alledges

    def interval(self, index):
        if index < 0:
            index += len(self)

        low = self._fXaxis._fXmin
        high = self._fXaxis._fXmax
        if index == 0:
            return (float("-inf"), low)
        elif index == len(self) - 1:
            return (high, float("inf"))
        elif len(getattr(self._fXaxis, "_fXbins", [])) == self._fXaxis._fNbins + 1:
            return (self._fXaxis._fXbins[index - 1], self._fXaxis._fXbins[index])
        else:
            norm = (high - low) / self._fXaxis._fNbins
            return (index - 1)*norm + low, index*norm + low

    @property
    def xlabels(self):
        if getattr(self._fXaxis, "_fLabels", None) is None:
            return None
        else:
            return [str(x) for x in self._fXaxis._fLabels]

    def show(self, width=80, minimum=None, maximum=None, stream=sys.stdout):
        if minimum is None:
            minimum = min(self)
            if minimum < 0:
                minimum *= 1.05
            else:
                minimum = 0

        if maximum is None:
            maximum = max(self) * 1.05

        if maximum <= minimum:
            average = (minimum + maximum) / 2.0
            minimum = average - 0.5
            maximum = average + 0.5

        if self.xlabels is None:
            intervals = ["[{0:<.5g}, {1:<.5g})".format(l, h) for l, h in [self.interval(i) for i in range(len(self))]]
            intervals[-1] = intervals[-1][:-1] + "]"   # last interval is closed on top edge
        else:
            intervals = ["(underflow)"] + [self.xlabels[i] if i < len(self.xlabels) else self.interval(i+1) for i in range(self.numbins)] + ["(overflow)"]

        intervalswidth = max(len(x) for x in intervals)

        values = ["{0:<.5g}".format(float(x)) for x in self]
        valueswidth = max(len(x) for x in values)

        minimumtext = "{0:<.5g}".format(minimum)
        maximumtext = "{0:<.5g}".format(maximum)

        plotwidth = max(len(minimumtext) + len(maximumtext), width - (intervalswidth + 1 + valueswidth + 1 + 2))
        scale = minimumtext + " "*(plotwidth + 2 - len(minimumtext) - len(maximumtext)) + maximumtext

        norm = float(plotwidth) / float(maximum - minimum)
        zero = int(round((0.0 - minimum)*norm))
        line = numpy.empty(plotwidth, dtype=numpy.uint8)

        formatter = "{0:<%s} {1:<%s} |{2}|" % (intervalswidth, valueswidth)
        line[:] = ord("-")
        if minimum != 0 and 0 <= zero < plotwidth:
            line[zero] = ord("+")
        capstone = " " * (intervalswidth + 1 + valueswidth + 1) + "+" + str(line.tostring().decode("ascii")) + "+"

        out = [" "*(intervalswidth + valueswidth + 2) + scale]
        out.append(capstone)
        for interval, value, x in zip(intervals, values, self):
            line[:] = ord(" ")

            pos = int(round((x - minimum)*norm))
            if x < 0:
                line[pos:zero] = ord("*")
            else:
                line[zero:pos] = ord("*")

            if minimum != 0 and 0 <= zero < plotwidth:
                line[zero] = ord("|")

            out.append(formatter.format(interval, value, str(line.tostring().decode("ascii"))))

        out.append(capstone)
        out = "\n".join(out)
        if stream is None:
            return out
        else:
            stream.write(out)
            stream.write("\n")

    def pandas(self, underflow=True, overflow=True, variance=True):

        if not underflow and not overflow:
            s = slice(1,-1)
        elif not underflow:
            s = slice(1, None)
        elif not overflow:
            s = slice(None,-1)
        else:
            s = slice(None)

        def get_name(obj):
            if getattr(obj, "_fTitle", b"") == b"":
                return None
            else:
                return obj._fTitle.decode("utf-8", "ignore")

        allbins = self.allbins
        if type(allbins) != tuple:
            allbins = (allbins,)

        dim = len(allbins)

        if dim == 1:
            index_name_objects = (self,)
        elif dim == 2:
            index_name_objects = (self._fXaxis, self._fYaxis)
        elif dim == 3:
            index_name_objects = (self._fXaxis, self._fYaxis, self._fZaxis)
        else:
            raise NotImplementedError

        data = {"count" : self.allvalues[(s,)*dim].flatten()}
        columns = ["count"]

        if variance:
            data["variance"] = self.allvariances[(s,)*dim].flatten()
            columns += ["variance"]

        import pandas

        intervals = ((pandas.Interval(a,b, closed="left") for a, b in bins[s]) for bins in allbins)
        names = (get_name(obj) for obj in index_name_objects)
        index = pandas.MultiIndex.from_product(intervals, names=names)

        return pandas.DataFrame(index=index, data=data, columns=columns)

    def physt(self):
        import physt.binnings
        import physt.histogram1d
        freq = numpy.array(self.allvalues, dtype=getattr(self, "_dtype", numpy.dtype(numpy.float64)).newbyteorder("="))
        if getattr(self._fXaxis, "_fXbins", None):
            binning = physt.binnings.NumpyBinning(numpy.array(self._fXaxis._fXbins))
        else:
            low = self._fXaxis._fXmin
            high = self._fXaxis._fXmax
            binwidth = (high - low) / self._fXaxis._fNbins
            binning = physt.binnings.FixedWidthBinning(binwidth, bin_count=self._fXaxis._fNbins, min=low)
        return physt.histogram1d.Histogram1D(
            binning,
            frequencies=freq[1:-1],
            underflow=freq[0],
            overflow=freq[-1],
            name=getattr(self, "_fTitle", b"").decode("utf-8", "ignore"))

    def hepdata(self, independent={"name": None, "units": None}, dependent={"name": "counts", "units": None}, qualifiers=[], yamloptions={}):
        if independent["name"] is None and getattr(self, "_fTitle", b""):
            independent = dict(independent)
            if isinstance(self._fTitle, bytes):
                independent["name"] = self._fTitle.decode("utf-8", "ignore")
            else:
                independent["name"] = self._fTitle

        if getattr(self._fXaxis, "_fXbins", None):
            independent_values = [{"low": float(low), "high": float(high)} for low, high in zip(self._fXaxis._fXbins[:-1], self._fXaxis._fXbins[1:])]
        else:
            low = self._fXaxis._fXmin
            high = self._fXaxis._fXmax
            norm = (high - low) / self._fXaxis._fNbins
            independent_values = [{"low": float(i*norm + low), "high": float((i + 1)*norm + low)} for i in range(self.numbins)]

        if getattr(self, "_fSumw2", None):
            dependent_values = [{"value": float(value), "errors": [{"symerror": math.sqrt(variance), "label": "stat"}]} for value, variance in zip(self.values, self.variances)]
        else:
            dependent_values = [{"value": float(value), "errors": [{"symerror": math.sqrt(value), "label": "stat"}]} for value in self.values]

        out = {"independent_variables": [{"header": independent, "values": independent_values}], "dependent_variables": [{"header": dependent, "qualifiers": qualifiers, "values": dependent_values}]}

        if yamloptions is None:
            return out
        else:
            import yaml
            return yaml.dump(out, **yamloptions)

def _histtype(content):
    if issubclass(content.dtype.type, (numpy.bool_, numpy.bool)):
        return b"TH1C", content.astype(">i1")
    elif issubclass(content.dtype.type, numpy.int8):
        return b"TH1C", content.astype(">i1")
    elif issubclass(content.dtype.type, numpy.uint8) and content.max() <= numpy.iinfo(numpy.int8).max:
        return b"TH1C", content.astype(">i1")
    elif issubclass(content.dtype.type, numpy.uint8):
        return b"TH1S", content.astype(">i2")
    elif issubclass(content.dtype.type, numpy.int16):
        return b"TH1S", content.astype(">i2")
    elif issubclass(content.dtype.type, numpy.uint16) and content.max() <= numpy.iinfo(numpy.int16).max:
        return b"TH1S", content.astype(">i2")
    elif issubclass(content.dtype.type, numpy.uint16):
        return b"TH1I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.int32):
        return b"TH1I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.uint32) and content.max() <= numpy.iinfo(numpy.int32).max:
        return b"TH1I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.integer) and numpy.iinfo(numpy.int32).min <= content.min() and content.max() <= numpy.iinfo(numpy.int32).max:
        return b"TH1I", content.astype(">i4")
    elif issubclass(content.dtype.type, numpy.float32):
        return b"TH1F", content.astype(">f4")
    else:
        return b"TH1D", content.astype(">f8")

def from_numpy(histogram):
    content, edges = histogram[:2]

    class TH1(Methods, list):
        pass

    class TAxis(object):
        def __init__(self, edges):
            self._fNbins = len(edges) - 1
            self._fXmin = edges[0]
            self._fXmax = edges[-1]
            if numpy.array_equal(edges, numpy.linspace(self._fXmin, self._fXmax, len(edges), dtype=edges.dtype)):
                self._fXbins = numpy.array([], dtype=">f8")
            else:
                self._fXbins = edges.astype(">f8")

    out = TH1.__new__(TH1)
    out._fXaxis = TAxis(edges)

    centers = (edges[:-1] + edges[1:]) / 2.0
    out._fEntries = out._fTsumw = out._fTsumw2 = content.sum()
    out._fTsumwx = (content * centers).sum()
    out._fTsumwx2 = (content * centers**2).sum()

    if len(histogram) >= 3:
        out._fTitle = histogram[2]
    else:
        out._fTitle = b""

    out._classname, content = _histtype(content)

    valuesarray = numpy.empty(len(content) + 2, dtype=content.dtype)
    valuesarray[1:-1] = content
    valuesarray[0] = 0
    valuesarray[-1] = 0

    out.extend(valuesarray)
    # variances equal the entries
    out._fSumw2 = valuesarray

    return out

def from_pandas(histogram):
    import pandas

    histogram = histogram.sort_index(ascending=True, inplace=False)
    if not histogram.index.is_non_overlapping_monotonic:
        raise ValueError("intervals overlap; cannot form a histogram")

    sparse = histogram.index[numpy.isfinite(histogram.index.left) & numpy.isfinite(histogram.index.right)]
    if (sparse.right[:-1] == sparse.left[1:]).all():
        dense = sparse
    else:
        pairs = numpy.empty(len(sparse) * 2, dtype=numpy.float64)
        pairs[::2] = sparse.left
        pairs[1::2] = sparse.right
        nonempty = numpy.empty(len(pairs), dtype=numpy.bool_)
        nonempty[:-1] = (pairs[1:] != pairs[:-1])
        nonempty[-1] = True
        dense = pandas.IntervalIndex.from_breaks(pairs[nonempty], closed="left")

    densehist = pandas.DataFrame(index=dense.left).join(histogram.reindex(histogram.index.left))
    densehist.fillna(0, inplace=True)

    underflowhist = histogram[numpy.isinf(histogram.index.left)]
    overflowhist = histogram[numpy.isinf(histogram.index.right)]

    content = numpy.array(densehist["count"])

    sumw2 = numpy.empty(len(content) + 2, dtype=numpy.float64)
    if "variance" in densehist.columns:
        sumw2source = "variance"
    else:
        sumw2source = "count"
    sumw2[1:-1] = densehist[sumw2source]
    if len(underflowhist) == 0:
        sumw2[0] = 0
    else:
        sumw2[0] = underflowhist[sumw2source]
    if len(overflowhist) == 0:
        sumw2[-1] = 0
    else:
        sumw2[-1] = overflowhist[sumw2source]

    edges = numpy.empty(len(densehist) + 1, dtype=numpy.float64)
    edges[:-1] = dense.left
    edges[-1] = dense.right[-1]

    class TH1(Methods, list):
        pass

    class TAxis(object):
        def __init__(self, fNbins, fXmin, fXmax):
            self._fNbins = fNbins
            self._fXmin = fXmin
            self._fXmax = fXmax

    out = TH1.__new__(TH1)
    out._fXaxis = TAxis(len(edges) - 1, edges[0], edges[-1])
    out._fXaxis._fXbins = edges

    centers = (edges[:-1] + edges[1:]) / 2.0
    out._fEntries = content.sum()
    out._fTsumw = content.sum()
    out._fTsumw2 = sumw2.sum()
    out._fTsumwx = (content * centers).sum()
    out._fTsumwx2 = (content * centers**2).sum()

    if histogram.index.name is None:
        out._fTitle = b""
    elif isinstance(histogram.index.name, bytes):
        out._fTitle = histogram.index.name
    else:
        out._fTitle = histogram.index.name.encode("utf-8", "ignore")

    out._classname, content = _histtype(content)

    valuesarray = numpy.empty(len(content) + 2, dtype=content.dtype)
    valuesarray[1:-1] = content
    if len(underflowhist) == 0:
        valuesarray[0] = 0
    else:
        valuesarray[0] = underflowhist["count"]
    if len(overflowhist) == 0:
        valuesarray[-1] = 0
    else:
        valuesarray[-1] = overflowhist["count"]

    out.extend(valuesarray)

    return out

def from_physt(histogram):
    import physt.binnings
    import physt.histogram1d

    class TH1(Methods, list):
        pass

    class TAxis(object):
        def __init__(self, fNbins, fXmin, fXmax):
            self._fNbins = fNbins
            self._fXmin = fXmin
            self._fXmax = fXmax

    out = TH1.__new__(TH1)

    if isinstance(histogram.binning, physt.binnings.FixedWidthBinning):
        out._fXaxis = TAxis(histogram.binning.bin_count,
                            histogram.binning.first_edge,
                            histogram.binning.last_edge)
    elif isinstance(histogram.binning, physt.binnings.NumpyBinning):
        out._fXaxis = TAxis(histogram.binning.bin_count,
                            histogram.binning.first_edge,
                            histogram.binning.last_edge)
        if not histogram.binning.is_regular():
            out._fXaxis._fXbins = histogram.binning.numpy_bins.astype(">f8")
    else:
        raise NotImplementedError(histogram.binning)

    centers = histogram.bin_centers
    content = histogram.frequencies

    out._fSumw2 = [0] + list(histogram.errors2) + [0]

    mean = histogram.mean()
    variance = histogram.variance()
    out._fEntries = content.sum()   # is there a #entries independent of weights?
    out._fTsumw = content.sum()
    out._fTsumw2 = histogram.errors2.sum()
    if mean is None:
        out._fTsumwx = (content * centers).sum()
    else:
        out._fTsumwx = mean * out._fTsumw
    if mean is None or variance is None:
        out._fTsumwx2 = (content * centers**2).sum()
    else:
        out._fTsumwx2 = (mean**2 + variance) * out._fTsumw2

    if histogram.name is not None:
        out._fTitle = histogram.name
    else:
        out._fTitle = b""

    out._classname, content = _histtype(content)

    valuesarray = numpy.empty(len(content) + 2, dtype=content.dtype)
    valuesarray[1:-1] = content
    valuesarray[0] = histogram.underflow
    valuesarray[-1] = histogram.overflow

    out.extend(valuesarray)

    return out



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

from uproot3_methods.classes.TVector2 import TVector2, TVector2Array
from uproot3_methods.classes.TVector3 import TVector3, TVector3Array
from uproot3_methods.classes.TLorentzVector import TLorentzVector, TLorentzVectorArray, PtEtaPhiMassLorentzVector

# convenient access to the version number
from uproot3_methods.version import __version__



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import math
import numbers

import awkward0.array.jagged
import awkward0.util

import uproot3_methods.base
import uproot3_methods.common.TVector

class Common(object):
    def dot(self, other):
        out = self.x*other.x
        out = out + self.y*other.y
        out = out + self.z*other.z
        return out

    def _cross(self, other):
        return (self.y*other.z - self.z*other.y,
                self.z*other.x - self.x*other.z,
                self.x*other.y - self.y*other.x)

    @property
    def cottheta(self):
        out = self.rho
        out /= self.z
        return out

    def _rotate_axis(self, axis, angle):
        u = axis.unit
        c = self.awkward0.numpy.cos(angle)
        s = self.awkward0.numpy.sin(angle)
        c1 = 1 - c

        x = (c + u.x**2 * c1) * self.x + (u.x * u.y * c1 - u.z * s) * self.y + (u.x * u.z * c1 + u.y * s) * self.z
        y = (u.x * u.y * c1 + u.z * s) * self.x + (c + u.y**2 * c1) * self.y + (u.y * u.z * c1 - u.x * s) * self.z
        z = (u.x * u.z * c1 - u.y * s) * self.x + (u.y * u.z * c1 + u.x * s) * self.y + (c + u.z**2 * c1) * self.z

        return x, y, z

    def _rotate_euler(self, phi, theta, psi):
        # Rotate Z (phi)
        c1 = self.awkward0.numpy.cos(phi)
        s1 = self.awkward0.numpy.sin(phi)
        c2 = self.awkward0.numpy.cos(theta)
        s2 = self.awkward0.numpy.sin(theta)
        c3 = self.awkward0.numpy.cos(psi)
        s3 = self.awkward0.numpy.sin(psi)

        # Rotate Y (theta)
        fzx2 = -s2*c1
        fzy2 =  s2*s1
        fzz2 =  c2

        # Rotate Z (psi)
        fxx3 =  c3*c2*c1 - s3*s1
        fxy3 = -c3*c2*s1 - s3*c1
        fxz3 =  c3*s2
        fyx3 =  s3*c2*c1 + c3*s1
        fyy3 = -s3*c2*s1 + c3*c1
        fyz3 =  s3*s2

        # Transform v
        x = fxx3*self.x + fxy3*self.y + fxz3*self.z
        y = fyx3*self.x + fyy3*self.y + fyz3*self.z
        z = fzx2*self.x + fzy2*self.y + fzz2*self.z

        return x, y, z

    def rotatex(self, angle):
        return self.rotate_axis(TVector3(1.0, 0.0, 0.0), angle)

    def rotatey(self, angle):
        return self.rotate_axis(TVector3(0.0, 1.0, 0.0), angle)

    def rotatez(self, angle):
        return self.rotate_axis(TVector3(0.0, 0.0, 1.0), angle)

class ArrayMethods(Common, uproot3_methods.common.TVector.ArrayMethods, uproot3_methods.base.ROOTMethods):
    def _initObjectArray(self, table):
        self.awkward0.ObjectArray.__init__(self, table, lambda row: TVector3(row["fX"], row["fY"], row["fZ"]))

    def __awkward_serialize__(self, serializer):
        self._valid()
        x, y, z = self.x, self.y, self.z
        return serializer.encode_call(
            ["uproot3_methods.classes.TVector3", "TVector3Array", "from_cartesian"],
            serializer(x, "TVector3Array.x"),
            serializer(y, "TVector3Array.y"),
            serializer(z, "TVector3Array.z"))

    @property
    def x(self):
        return self["fX"]

    @property
    def y(self):
        return self["fY"]

    @property
    def z(self):
        return self["fZ"]

    def cross(self, other):
        x, y, z = self._cross(other)
        out = self.empty_like()
        out["fX"] = x
        out["fY"] = y
        out["fZ"] = z
        return out

    @property
    def theta(self):
        return self.awkward0.numpy.arctan2(self.rho, self.z)

    def rotate_axis(self, axis, angle):
        x, y, z = self._rotate_axis(axis, angle)
        out = self.empty_like()
        out["fX"] = x
        out["fY"] = y
        out["fZ"] = z
        return out

    def rotate_euler(self, phi=0, theta=0, psi=0):
        x, y, z = self._rotate_euler(phi, theta, psi)
        out = self.empty_like()
        out["fX"] = x
        out["fY"] = y
        out["fZ"] = z
        return out

    def sum(self):
        if isinstance(self, self.awkward0.JaggedArray):
            return TVector3Array.from_cartesian(self.x.sum(), self.y.sum(), self.z.sum())
        else:
            return TVector3(self.x.sum(), self.y.sum(), self.z.sum())

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if "out" in kwargs:
            raise NotImplementedError("in-place operations not supported")

        if method != "__call__":
            return NotImplemented

        inputs = list(inputs)
        for i in range(len(inputs)):
            if isinstance(inputs[i], self.awkward0.numpy.ndarray) and inputs[i].dtype == self.awkward0.numpy.dtype(object) and len(inputs[i]) > 0:
                idarray = self.awkward0.numpy.frombuffer(inputs[i], dtype=self.awkward0.numpy.uintp)
                if (idarray == idarray[0]).all():
                    inputs[i] = inputs[i][0]

        if ufunc is self.awkward0.numpy.add or ufunc is self.awkward0.numpy.subtract:
            if not all(isinstance(x, (ArrayMethods, Methods)) for x in inputs):
                raise TypeError("(arrays of) TVector3 can only be added to/subtracted from other (arrays of) TVector3")
            out = self.empty_like()
            out["fX"] = getattr(ufunc, method)(*[x.x for x in inputs], **kwargs)
            out["fY"] = getattr(ufunc, method)(*[x.y for x in inputs], **kwargs)
            out["fZ"] = getattr(ufunc, method)(*[x.z for x in inputs], **kwargs)
            return out

        elif ufunc is self.awkward0.numpy.power and len(inputs) >= 2 and isinstance(inputs[1], (numbers.Number, self.awkward0.numpy.number)):
            if inputs[1] == 2:
                return self.mag2
            else:
                return self.mag2**(0.5*inputs[1])

        elif ufunc is self.awkward0.numpy.absolute:
            return self.mag

        else:
            return super(ArrayMethods, self).__array_ufunc__(ufunc, method, *inputs, **kwargs)

JaggedArrayMethods = ArrayMethods.mixin(ArrayMethods, awkward0.JaggedArray)

class Methods(Common, uproot3_methods.common.TVector.Methods, uproot3_methods.base.ROOTMethods):
    _arraymethods = ArrayMethods

    @property
    def x(self):
        return self._fX

    @property
    def y(self):
        return self._fY

    @property
    def z(self):
        return self._fZ

    def __repr__(self):
        return "TVector3({0:.5g}, {1:.5g}, {2:.5g})".format(self.x, self.y, self.z)

    def __str__(self):
        return repr(self)

    def __eq__(self, other):
        return isinstance(other, Methods) and self.x == other.x and self.y == other.y and self.z == other.z

    def _scalar(self, operator, scalar, reverse=False):
        if not isinstance(scalar, (numbers.Number, self.awkward0.numpy.number)):
            raise TypeError("cannot {0} a TVector3 with a {1}".format(operator.__name__, type(scalar).__name__))
        if reverse:
            return TVector3(operator(scalar, self.x), operator(scalar, self.y), operator(scalar, self.z))
        else:
            return TVector3(operator(self.x, scalar), operator(self.y, scalar), operator(self.z, scalar))

    def _vector(self, operator, vector, reverse=False):
        if not isinstance(vector, Methods):
            raise TypeError("cannot {0} a TVector3 with a {1}".format(operator.__name__, type(vector).__name__))
        if reverse:
            return TVector3(operator(vector.x, self.x), operator(vector.y, self.y), operator(vector.z, self.z))
        else:
            return TVector3(operator(self.x, vector.x), operator(self.y, vector.y), operator(self.z, vector.z))

    def _unary(self, operator):
        return TVector3(operator(self.x), operator(self.y), operator(self.z))

    def cross(self, other):
        x, y, z = self._cross(other)
        return TVector3(x, y, z)

    def angle(self, other):
        denominator = math.sqrt(self.dot(self) * other.dot(other))
        if denominator == 0:
            # one of the vector is null
            return 0.
        cos_angle = self.dot(other) / denominator
        if cos_angle > 1:
            cos_angle = 1
        elif cos_angle < -1:
            cos_angle = -1
        return math.acos(cos_angle)

    @property
    def theta(self):
        return math.atan2(self.rho, self.z)

    def rotate_axis(self, axis, angle):
        x, y, z = self._rotate_axis(axis, angle)
        return TVector3(x, y, z)

    def rotate_euler(self, phi=0, theta=0, psi=0):
        return TVector3(x, y, z)

class TVector3Array(ArrayMethods, uproot3_methods.base.ROOTMethods.awkward0.ObjectArray):

    def __init__(self, x, y, z):
        if isinstance(x, awkward0.array.jagged.JaggedArray) or isinstance(y, awkward0.array.jagged.JaggedArray) or isinstance(z, awkward0.array.jagged.JaggedArray):
            raise TypeError("TVector3Array constructor arguments must not be jagged; use TVector3.from_cartesian for jaggedness-handling")
        self._initObjectArray(self.awkward0.Table())
        self["fX"] = x
        self["fY"] = y
        self["fZ"] = z

    @classmethod
    def origin(cls, shape, dtype=None):
        if dtype is None:
            dtype = cls.awkward0.numpy.float64
        return cls(cls.awkward0.numpy.zeros(shape, dtype=dtype),
                   cls.awkward0.numpy.zeros(shape, dtype=dtype),
                   cls.awkward0.numpy.zeros(shape, dtype=dtype))

    @classmethod
    def origin_like(cls, array):
        return array * 0.0

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_cartesian(cls, x, y, z):
        return cls(x, y, z)

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_spherical(cls, r, theta, phi):
        return cls(r * cls.awkward0.numpy.sin(theta) * cls.awkward0.numpy.cos(phi),
                   r * cls.awkward0.numpy.sin(theta) * cls.awkward0.numpy.sin(phi),
                   r * cls.awkward0.numpy.cos(theta))

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_cylindrical(cls, rho, phi, z):
        return cls(rho * cls.awkward0.numpy.cos(phi), rho * cls.awkward0.numpy.sin(phi),z)

    @property
    def x(self):
        return self["fX"]

    @x.setter
    def x(self, value):
        self["fX"] = value

    @property
    def y(self):
        return self["fY"]

    @y.setter
    def y(self, value):
        self["fY"] = value

    @property
    def z(self):
        return self["fZ"]

    @z.setter
    def z(self, value):
        self["fZ"] = value

class TVector3(Methods):
    def __init__(self, x, y, z):
        self._fX = float(x)
        self._fY = float(y)
        self._fZ = float(z)

    @classmethod
    def origin(cls):
        return cls(0.0, 0.0, 0.0)

    @classmethod
    def from_spherical(cls, r, theta, phi):
        return cls(r * math.sin(theta) * math.cos(phi),
                   r * math.sin(theta) * math.sin(phi),
                   r * math.cos(theta))

    @classmethod
    def from_cylindrical(cls, rho, phi, z):
        return cls(rho * math.cos(phi),
                   rho * math.sin(phi),
                   z)

    @property
    def x(self):
        return self._fX

    @x.setter
    def x(self, value):
        self._fX = value

    @property
    def y(self):
        return self._fY

    @y.setter
    def y(self, value):
        self._fY = value

    @property
    def z(self):
        return self._fZ

    @z.setter
    def z(self, value):
        self._fZ = value



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/uproot3-methods/blob/master/LICENSE

import math
import numbers

import awkward0.array.jagged
import awkward0.util

import uproot3_methods.common.TVector
import uproot3_methods.base

class Common(object):
    def dot(self, other):
        out = self.x*other.x
        out = out + self.y*other.y
        return out

    def _rotate(self, angle):
        c = self.awkward0.numpy.cos(angle)
        s = self.awkward0.numpy.sin(angle)
        x = self.x*c - self.y*s
        y = self.x*s + self.y*c
        return x, y

class ArrayMethods(Common, uproot3_methods.common.TVector.ArrayMethods, uproot3_methods.base.ROOTMethods):
    def _initObjectArray(self, table):
        self.awkward0.ObjectArray.__init__(self, table, lambda row: TVector2(row["fX"], row["fY"]))

    def __awkward_serialize__(self, serializer):
        self._valid()
        x, y = self.x, self.y
        return serializer.encode_call(
            ["uproot3_methods.classes.TVector2", "TVector2Array", "from_cartesian"],
            serializer(x, "TVector3Array.x"),
            serializer(y, "TVector3Array.y"))

    @property
    def x(self):
        return self["fX"]

    @property
    def y(self):
        return self["fY"]

    def rotate(self, angle):
        x, y = self._rotate(angle)
        out = self.empty_like()
        out["fX"] = x
        out["fY"] = y
        return out

    def sum(self):
        if isinstance(self, self.awkward0.JaggedArray):
            return TVector2Array.from_cartesian(self.x.sum(), self.y.sum())
        else:
            return TVector2(self.x.sum(), self.y.sum())

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if "out" in kwargs:
            raise NotImplementedError("in-place operations not supported")

        if method != "__call__":
            return NotImplemented

        inputs = list(inputs)
        for i in range(len(inputs)):
            if isinstance(inputs[i], self.awkward0.numpy.ndarray) and inputs[i].dtype == self.awkward0.numpy.dtype(object) and len(inputs[i]) > 0:
                idarray = self.awkward0.numpy.frombuffer(inputs[i], dtype=self.awkward0.numpy.uintp)
                if (idarray == idarray[0]).all():
                    inputs[i] = inputs[i][0]

        if ufunc is self.awkward0.numpy.add or ufunc is self.awkward0.numpy.subtract:
            if not all(isinstance(x, (ArrayMethods, Methods)) for x in inputs):
                raise TypeError("(arrays of) TVector2 can only be added to/subtracted from other (arrays of) TVector2")
            out = self.empty_like()
            out["fX"] = getattr(ufunc, method)(*[x.x for x in inputs], **kwargs)
            out["fY"] = getattr(ufunc, method)(*[x.y for x in inputs], **kwargs)
            return out

        elif ufunc is self.awkward0.numpy.power and len(inputs) >= 2 and isinstance(inputs[1], (numbers.Number, self.awkward0.numpy.number)):
            if inputs[1] == 2:
                return self.mag2
            else:
                return self.mag2**(0.5*inputs[1])

        elif ufunc is self.awkward0.numpy.absolute:
            return self.mag

        else:
            return super(ArrayMethods, self).__array_ufunc__(ufunc, method, *inputs, **kwargs)

JaggedArrayMethods = ArrayMethods.mixin(ArrayMethods, awkward0.JaggedArray)

class Methods(Common, uproot3_methods.common.TVector.Methods, uproot3_methods.base.ROOTMethods):
    _arraymethods = ArrayMethods

    @property
    def x(self):
        return self._fX

    @property
    def y(self):
        return self._fY

    def __repr__(self):
        return "TVector2({0:.5g}, {1:.5g})".format(self.x, self.y)

    def __str__(self):
        return repr(self)

    def __eq__(self, other):
        return isinstance(other, Methods) and self.x == other.x and self.y == other.y

    def _scalar(self, operator, scalar, reverse=False):
        if not isinstance(scalar, (numbers.Number, self.awkward0.numpy.number)):
            raise TypeError("cannot {0} a TVector2 with a {1}".format(operator.__name__, type(scalar).__name__))
        if reverse:
            return TVector2(operator(scalar, self.x), operator(scalar, self.y))
        else:
            return TVector2(operator(self.x, scalar), operator(self.y, scalar))

    def _vector(self, operator, vector, reverse=False):
        if not isinstance(vector, Methods):
            raise TypeError("cannot {0} a TVector2 with a {1}".format(operator.__name__, type(vector).__name__))
        if reverse:
            return TVector2(operator(vector.x, self.x), operator(vector.y, self.y))
        else:
            return TVector2(operator(self.x, vector.x), operator(self.y, vector.y))

    def _unary(self, operator):
        return TVector2(operator(self.x), operator(self.y))

    def rotate(self, angle):
        x, y = self._rotate(angle)
        return TVector2(x, y)

class TVector2Array(ArrayMethods, uproot3_methods.base.ROOTMethods.awkward0.ObjectArray):

    def __init__(self, x, y):
        if isinstance(x, awkward0.array.jagged.JaggedArray) or isinstance(y, awkward0.array.jagged.JaggedArray):
            raise TypeError("TVector2Array constructor arguments must not be jagged; use TVector2.from_cartesian for jaggedness-handling")
        self._initObjectArray(self.awkward0.Table())
        self["fX"] = x
        self["fY"] = y

    @classmethod
    def origin(cls, shape, dtype=None):
        if dtype is None:
            dtype = cls.awkward0.numpy.float64
        return cls(cls.awkward0.numpy.zeros(shape, dtype=dtype), cls.awkward0.numpy.zeros(shape, dtype=dtype))

    @classmethod
    def origin_like(cls, array):
        return array * 0.0

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_cartesian(cls, x, y):
        return cls(x, y)

    @classmethod
    @awkward0.util.wrapjaggedmethod(JaggedArrayMethods)
    def from_polar(cls, rho, phi):
        return cls(rho * cls.awkward0.numpy.cos(phi),
                   rho * cls.awkward0.numpy.sin(phi))

    @property
    def x(self):
        return self["fX"]

    @x.setter
    def x(self, value):
        self["fX"] = value

    @property
    def y(self):
        return self["fY"]

    @y.setter
    def y(self, value):
        self["fY"] = value

class TVector2(Methods):
    def __init__(self, x, y):
        self._fX = float(x)
        self._fY = float(y)

    @classmethod
    def origin(cls):
        return cls(0.0, 0.0)

    @classmethod
    def from_polar(cls, rho, phi):
        return cls(rho * math.cos(phi),
                   rho * math.sin(phi))

    @property
    def x(self):
        return self._fX

    @x.setter
    def x(self, value):
        self._fX = value

    @property
    def y(self):
        return self._fY

    @y.setter
    def y(self, value):
        self._fY = value
